from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.platform import Platform
from fsgs.platforms.loader import CartridgePlatformLoader

"""
Some information from https://en.wikipedia.org/wiki/TurboGrafx-16:

All PC Engine hardware outputs video in NTSC format, including the European
TurboGrafx; it generates a PAL-compatible video signal by using a chroma
encoder chip not found in any other system in the series.

X (Horizontal) Resolution: variable, maximum of 565 (programmable to 282,
377 or 565 pixels, or as 5.3693175 MHz, 7.15909 MHz, and 10.738635 MHz pixel
dot clock). Taking into consideration overscan limitations of CRT televisions
at the time, the horizontal resolutions were realistically limited to
something a bit less than what the system was actually capable of.
Consequently, most game developers limited their games to either 256, 352,
or 512 pixels in display width for each of the three modes.[23]

Y (Vertical) Resolution: variable, maximum of 242 (programmable in increments
of 1 scanline). It is possible to achieve an interlaced "mode" with a maximum
vertical resolution of 484 scanlines by alternating between the two different
vertical resolution modes used by the system. However, it is unknown, at this
time, if this interlaced resolution is compliant with (and hence displayed
correctly on) NTSC televisions.

The majority of TurboGrafx-16 games use 256×239
"""

TG16_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "turbografx16",
}


class TurboGrafx16Platform(Platform):
    PLATFORM_ID = "tg16"
    PLATFORM_NAME = "TurboGrafx-16"

    def __init__(self):
        super().__init__()

    def driver(self, fsgc):
        return TurboGrafx16MednafenDriver(fsgc)

    def loader(self, fsgc):
        return TurboGrafx16Loader(fsgc)


class TurboGrafx16Loader(CartridgePlatformLoader):
    pass


class TurboGrafx16MednafenDriver(MednafenDriver):
    PORTS = [
        {"description": "Input Port 1", "types": [TG16_CONTROLLER]},
        {"description": "Input Port 2", "types": [TG16_CONTROLLER]},
    ]

    def __init__(self, fsgc):
        super().__init__(fsgc)

    def game_video_par(self):
        size = self.game_video_size()
        return (4 / 3) / (size[0] / size[1])

    def game_video_size(self):
        # if self.is_pal():
        #     size = (256, 240)
        # else:
        #     size = (256, 240)
        size = (288, 232)
        return size

    def get_game_refresh_rate(self):
        return 59.94

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "1": "pce.input.port1.gamepad.i",
                "2": "pce.input.port1.gamepad.ii",
                "UP": "pce.input.port1.gamepad.up",
                "DOWN": "pce.input.port1.gamepad.down",
                "LEFT": "pce.input.port1.gamepad.left",
                "RIGHT": "pce.input.port1.gamepad.right",
                "SELECT": "pce.input.port1.gamepad.select",
                "RUN": "pce.input.port1.gamepad.run",
            }
        elif port == 1:
            return {
                "1": "pce.input.port2.gamepad.i",
                "2": "pce.input.port2.gamepad.ii",
                "UP": "pce.input.port2.gamepad.up",
                "DOWN": "pce.input.port2.gamepad.down",
                "LEFT": "pce.input.port2.gamepad.left",
                "RIGHT": "pce.input.port2.gamepad.right",
                "SELECT": "pce.input.port2.gamepad.select",
                "RUN": "pce.input.port2.gamepad.run",
            }

    def mednafen_system_prefix(self):
        return "pce"



import platform
import sys

from launcher.version import VERSION


def app_main():
    if "--help" in sys.argv:
        print(help_text)
        return
    print("FS-UAE Arcade {0}".format(VERSION))
    print(sys.argv)
    print(platform.uname())

    for i, arg in enumerate(sys.argv):
        if arg.startswith("--monitor="):
            sys.argv[i] = "--settings:monitor=" + arg[10:]
    try:
        import arcade.arcade_main

        arcade.arcade_main.main()
        print("main returned")

    finally:
        from fsbc.application import Application

        application = Application.instance()
        if application:
            print("calling Application stop")
            Application.get().stop()

        from fsbc.signal import Signal

        print("sending quit signal")
        Signal("quit").notify()


help_text = """\
FS-UAE Arcade {version}

Options:
  --fullscreen
  --window[=maximize]
  --monitor=left|right|middle-left|middle-right  (Requires fullscreen)
  --platform=<platform>                Open with chosen platform pre-selected
  --favorites                          Open with favorites pre-selected
  --disable-search                     Disable search function in Arcade UI

TODO: Add more documentation
""".format(
    version=VERSION
)



import os
import sys
import functools

# noinspection PyUnresolvedReferences
from fsbc.system import windows, macosx

# noinspection PyUnresolvedReferences
from fsbc.user import get_data_dir

# noinspection PyUnresolvedReferences
from .util import memoize


def cache(func):
    first_time = [True, None]

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if first_time[0]:
            first_time[1] = func(*args, **kwargs)
            first_time[0] = False
        return first_time[1]

    return wrapper


@cache
def get_lib_dir():
    lib_dir = os.environ.get("LIB_DIR", "").decode("UTF-8")
    if lib_dir:
        return unicode_path(lib_dir)
    lib_dir = os.path.join(os.getcwdu(), "lib")
    if os.path.exists(lib_dir):
        return lib_dir
    # raise RuntimeError("could not detect lib dir")
    return ""


EXCEPTION = "EXCEPTION"


@cache
def get_app_id():
    # FIXME
    return "fs-uae"
    # if windows or macosx:
    #     return "fs-uae"
    # else:
    #     return "fs-game-center"


@memoize
def get_app_data_dir(app=None):
    if not app:
        app = get_app_id()
    path = os.path.join(get_data_dir(), app)
    if not os.path.exists(path):
        os.makedirs(path)
    return path


@memoize
def get_app_config_dir(app=None):
    if not app:
        app = get_app_id()
    if windows:
        path = os.path.join(get_app_data_dir())
    elif macosx:
        path = os.path.join(get_home_dir(), "Library", "Preferences", app)
    else:
        path = os.path.join(get_home_dir(), ".config")
        path = os.environ.get("XDG_CONFIG_HOME", path)
        path = os.path.join(path, app)
        path = unicode_path(path)
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


def cause(exc, _cause):
    exc.__cause__ = _cause
    _cause.__traceback__ = sys.exc_info()[2]
    return exc


def encode_path(path):
    return path


def unicode_path(path):
    return path


def from_utf8_str(obj):
    if isinstance(obj, bytes):
        return obj.decode("UTF-8")
    return str(obj)


def to_utf8_str(obj):
    return obj


def utf8(obj):
    return unicode_safe(obj, "utf-8").encode("utf-8")


def utf8_safe(obj):
    return unicode_safe(obj, "utf-8").encode("utf-8")


def unicode_safe(obj, encoding="ASCII"):
    try:
        return str(obj)
    except Exception:
        pass
    try:
        return str(obj, encoding, "replace")
    except Exception:
        pass
    try:
        return str(str(obj), encoding, "replace")
    except Exception:
        # logger.exception("Error in unicode_safe")
        return "String returned from unicode_safe (problem logged)"


def normalize_path(path):
    path = os.path.normcase(os.path.normpath(path))
    return unicode_path(path)


# noinspection PyUnresolvedReferences
from .user import get_home_dir

# noinspection PyUnresolvedReferences
from .util import Version

# noinspection PyUnresolvedReferences
from .util import split_version

# noinspection PyUnresolvedReferences
from .util import compare_versions



# Copyright (c) 2008 Duncan Fordyce
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# taken from python irclib.. who took it from...
# Numeric table mostly stolen from the Perl IRC module (Net::IRC).
numeric_events = {
    "001": "welcome",
    "002": "yourhost",
    "003": "created",
    "004": "myinfo",
    "005": "featurelist",  # XXX
    "200": "tracelink",
    "201": "traceconnecting",
    "202": "tracehandshake",
    "203": "traceunknown",
    "204": "traceoperator",
    "205": "traceuser",
    "206": "traceserver",
    "207": "traceservice",
    "208": "tracenewtype",
    "209": "traceclass",
    "210": "tracereconnect",
    "211": "statslinkinfo",
    "212": "statscommands",
    "213": "statscline",
    "214": "statsnline",
    "215": "statsiline",
    "216": "statskline",
    "217": "statsqline",
    "218": "statsyline",
    "219": "endofstats",
    "221": "umodeis",
    "231": "serviceinfo",
    "232": "endofservices",
    "233": "service",
    "234": "servlist",
    "235": "servlistend",
    "241": "statslline",
    "242": "statsuptime",
    "243": "statsoline",
    "244": "statshline",
    "250": "luserconns",
    "251": "luserclient",
    "252": "luserop",
    "253": "luserunknown",
    "254": "luserchannels",
    "255": "luserme",
    "256": "adminme",
    "257": "adminloc1",
    "258": "adminloc2",
    "259": "adminemail",
    "261": "tracelog",
    "262": "endoftrace",
    "263": "tryagain",
    "265": "n_local",
    "266": "n_global",
    "300": "none",
    "301": "away",
    "302": "userhost",
    "303": "ison",
    "305": "unaway",
    "306": "nowaway",
    "311": "whoisuser",
    "312": "whoisserver",
    "313": "whoisoperator",
    "314": "whowasuser",
    "315": "endofwho",
    "316": "whoischanop",
    "317": "whoisidle",
    "318": "endofwhois",
    "319": "whoischannels",
    "321": "liststart",
    "322": "list",
    "323": "listend",
    "324": "channelmodeis",
    "329": "channelcreate",
    "331": "notopic",
    "332": "currenttopic",
    "333": "topicinfo",
    "341": "inviting",
    "342": "summoning",
    "346": "invitelist",
    "347": "endofinvitelist",
    "348": "exceptlist",
    "349": "endofexceptlist",
    "351": "version",
    "352": "whoreply",
    "353": "namreply",
    "361": "killdone",
    "362": "closing",
    "363": "closeend",
    "364": "links",
    "365": "endoflinks",
    "366": "endofnames",
    "367": "banlist",
    "368": "endofbanlist",
    "369": "endofwhowas",
    "371": "info",
    "372": "motd",
    "373": "infostart",
    "374": "endofinfo",
    "375": "motdstart",
    "376": "endofmotd",
    "377": "motd2",  # 1997-10-16 -- tkil
    "381": "youreoper",
    "382": "rehashing",
    "384": "myportis",
    "391": "time",
    "392": "usersstart",
    "393": "users",
    "394": "endofusers",
    "395": "nousers",
    "401": "nosuchnick",
    "402": "nosuchserver",
    "403": "nosuchchannel",
    "404": "cannotsendtochan",
    "405": "toomanychannels",
    "406": "wasnosuchnick",
    "407": "toomanytargets",
    "409": "noorigin",
    "411": "norecipient",
    "412": "notexttosend",
    "413": "notoplevel",
    "414": "wildtoplevel",
    "421": "unknowncommand",
    "422": "nomotd",
    "423": "noadmininfo",
    "424": "fileerror",
    "431": "nonicknamegiven",
    "432": "erroneusnickname",  # Thiss iz how its speld in thee RFC.
    "433": "nicknameinuse",
    "436": "nickcollision",
    "437": "unavailresource",  # "Nick temporally unavailable"
    "441": "usernotinchannel",
    "442": "notonchannel",
    "443": "useronchannel",
    "444": "nologin",
    "445": "summondisabled",
    "446": "usersdisabled",
    "451": "notregistered",
    "461": "needmoreparams",
    "462": "alreadyregistered",
    "463": "nopermforhost",
    "464": "passwdmismatch",
    "465": "yourebannedcreep",  # I love this one...
    "466": "youwillbebanned",
    "467": "keyset",
    "471": "channelisfull",
    "472": "unknownmode",
    "473": "inviteonlychan",
    "474": "bannedfromchan",
    "475": "badchannelkey",
    "476": "badchanmask",
    "477": "nochanmodes",  # "Channel doesn't support modes"
    "478": "banlistfull",
    "481": "noprivileges",
    "482": "chanoprivsneeded",
    "483": "cantkillserver",
    "484": "restricted",  # Connection is restricted
    "485": "uniqopprivsneeded",
    "491": "nooperhost",
    "492": "noservicehost",
    "501": "umodeunknownflag",
    "502": "usersdontmatch",
}

generated_events = [
    # Generated events
    "dcc_connect",
    "dcc_disconnect",
    "dccmsg",
    "disconnect",
    "ctcp",
    "ctcpreply",
]

protocol_events = [
    # IRC protocol events
    "error",
    "join",
    "kick",
    "mode",
    "part",
    "ping",
    "privmsg",
    "privnotice",
    "pubmsg",
    "pubnotice",
    "quit",
    "invite",
    "pong",
]

all_events = generated_events + protocol_events + list(numeric_events.values())




from fsui.qt import Qt, QDialog
from fsui.qt.helpers import QParent
from fsui.qt.windowbase import WindowBase


class Dialog(WindowBase(QDialog)):
    def __init__(self, parent=None, title=""):
        super().__init__(QParent(parent), title=title)

    def __del__(self):
        print("fsui.Dialog.__del__", self)

    def real_window(self):
        return self

    def show_modal(self):
        self.center_on_initial_show()
        # self.setWindowModality(Qt.WindowModal)
        return self.exec_()

    def show(self):
        self.center_on_initial_show()
        self.setWindowModality(Qt.WindowModal)
        return self.exec_()

    def end_modal(self, value):
        self.done(value)



import os

from fsbc.settings import Settings
from fsgs import Option
from fsgs.amiga.fsuae import FSUAE
from fsgs.amiga.launchhandler import LaunchHandler
from fsgs.drivers.gamedriver import GameDriver

AMIGA_JOYSTICK = {
    "type": "joystick",
    "description": "Joystick",
    "mapping_name": "amiga",
}
AMIGA_MOUSE = {
    "type": "mouse",
    "description": "Mouse",
    "mapping_name": "amiga_mouse",  # FIXME ...
}
CD32_CONTROLLER = {
    "type": "cd32 gamepad",
    "description": "CD32 Gamepad",
    "mapping_name": "cd32",
}
# FIXME: NO_DEVICE
AMIGA_PORTS = [
    {
        "description": "Joystick Port",
        "types": [AMIGA_JOYSTICK, AMIGA_MOUSE, CD32_CONTROLLER],
        "type_option": "joystick_port_1_mode",
        "device_option": "joystick_port_1",
    },
    {
        "description": "Mouse Port",
        "types": [AMIGA_JOYSTICK, AMIGA_MOUSE, CD32_CONTROLLER],
        "type_option": "joystick_port_0_mode",
        "device_option": "joystick_port_0",
    },
    # FIXME: ADDITIONAL PORTS...
]


class FSUAEAmigaDriver(GameDriver):
    PORTS = AMIGA_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.temp_config_file = None
        self.launch_handler = None

    def prepare(self):
        print("FSUAEAmigaDriver.prepare")

        # self.temp_dir = self.fsgc.temp_dir("amiga")

        # self.change_handler = GameChangeHandler(self.temp_dir)

        # self.firmware_dir = self.prepare_firmware("Amiga Firmware")
        # config = self.fsgc.config.copy()

        model = self.options[Option.AMIGA_MODEL]
        if model.startswith("CD32"):
            platform = "CD32"
        elif model == "CDTV":
            platform = "CDTV"
        else:
            platform = "Amiga"
        # name = Settings.get("config_name")
        # name = self.fsgc.game.name

        # uuid = Config.get("x_game_uuid")
        # uuid = None

        from fsgs.saves import SaveHandler

        save_state_handler = SaveHandler(self.fsgc, self.get_name(), platform)

        print(
            "[INPUT] joystick_port_1",
            self.options["joystick_port_1"],
            "->",
            self.ports[0].device_id or "none",
        )
        print(
            "[INPUT] joystick_port_0",
            self.options["joystick_port_0"],
            "->",
            self.ports[1].device_id or "none",
        )
        self.options["joystick_port_1"] = self.ports[0].device_id or "none"
        self.options["joystick_port_0"] = self.ports[1].device_id or "none"

        self.launch_handler = LaunchHandler(
            self.fsgc,
            self.get_name(),
            self.options,
            save_state_handler,
            temp_dir=self.cwd.path,
        )

        # self.change_handler.init(self.fsgc.get_game_state_dir(),
        #         ignore=["*.uss", "*.sdf"])

        # self.launch_handler.config["joystick_port_0"] = \
        #         self.inputs[1].device_id
        # self.launch_handler.config["joystick_port_1"] = \
        #         self.inputs[0].device_id

        if self.use_fullscreen():
            self.launch_handler.config["fullscreen"] = "1"
            if not self.launch_handler.config.get("fullscreen_mode", ""):
                # Check if fullscreen mode is overridden by temporary setting.
                if Settings.instance()["__fullscreen_mode"]:
                    self.launch_handler.config[
                        "fullscreen_mode"
                    ] = Settings.instance()["__fullscreen_mode"]
            if Settings.instance()["__arcade"]:
                # Remove window border when launched from FS-UAE Arcade in
                # order to reduce flickering
                self.launch_handler.config["window_border"] = "0"
                # Set fade out duration to 500, if not already specified.
                if not self.launch_handler.config.get("fade_out_duration", ""):
                    self.launch_handler.config["fade_out_duration"] = "500"
        else:
            self.launch_handler.config["fullscreen"] = "0"

        self.launch_handler.prepare()

    # def configure(self):
    #     print("AmigaGameHandler.configure")
    #
    #     #temp_dir = self.fsgc.temp.dir("amiga-config")
    #     #config_file = os.path.join(temp_dir, "amiga-config.fs-uae")
    #     #with open(config_file, "wb") as f:
    #     #    #self._configure_emulator(f)
    #     #    self.launch_handler.write_config(f)
    #     #    self.write_additional_config(f)#
    #     #
    #     ##self.args.extend(["--config", config_file])
    #     #self.args.extend([config_file])

    # def write_additional_config(self, f):
    #     #if self.get_option("fullscreen"):
    #     #    f.write("fullscreen = 1\n")
    #
    #     #if self.configure_vsync():
    #     #    f.write("video_sync = full\n")
    #     #else:
    #     #    f.write("video_sync = none\n")
    #
    #     if self.get_option("fsaa"):
    #         f.write("fsaa = {0}\n".format(str(self.get_option("fsaa"))))

    def run(self):
        print("FSUAEAmigaDriver.run, cwd =", self.cwd.path)
        config = self.launch_handler.create_config()
        self.emulator.process, self.temp_config_file = FSUAE.start_with_config(
            config, cwd=self.cwd.path
        )
        self.write_emulator_pid_file(
            self.emulator_pid_file(), self.emulator.process
        )

    def finish(self):
        print("FSUAEAmigaDriver.finish: Removing", self.temp_config_file)
        if self.temp_config_file is not None:
            try:
                os.remove(self.temp_config_file)
            except Exception:
                print("Could not remove config file", self.temp_config_file)

        self.launch_handler.update_changes()
        self.launch_handler.cleanup()
        # self.change_handler.update(self.fsgc.get_game_state_dir())

    def get_game_refresh_rate(self):
        # FIXME: Now assuming that all games are PAL / 50 Hz
        # - make configurable?
        return 50.0

    def get_supported_filters(self):
        supported = []
        # supported = [{
        #         "name": "2x",
        #         "gfx_filter": "none",
        #         "line_double": True,
        #         "lores": False,
        #     }, {
        #         "name": "none",
        #         "gfx_filter": "none",
        #         "line_double": False,
        #         "lores": False,
        #     }, {
        #         "name": "pal",
        #         "gfx_filter": "pal",
        #         "line_double": True,
        #         "lores": False,
        #     }, {
        #         "name": "scale2x",
        #         "gfx_filter": "scale2x",
        #         "line_double": False,
        #         "lores": True,
        #     }, {
        #         "name": "hq2x",
        #         "gfx_filter": "hq2x",
        #         "line_double": False,
        #         "lores": True,
        #     }
        # ]
        return supported

    # def set_input_options(self, f):
    #     # FIXME
    #     input_mapping = [{
    #         "1": "JOY2_FIRE_BUTTON",
    #         "2": "JOY2_2ND_BUTTON",
    #         "3": "JOY2_3RD_BUTTON",
    #         "left": "JOY2_LEFT",
    #         "right": "JOY2_RIGHT",
    #         "up": "JOY2_UP",
    #         "down": "JOY2_DOWN",
    #     }, {
    #         "1": "JOY1_FIRE_BUTTON",
    #         "2": "JOY1_2ND_BUTTON",
    #         "3": "JOY1_3RD_BUTTON",
    #         "left": "JOY1_LEFT",
    #         "right": "JOY1_RIGHT",
    #         "up": "JOY1_UP",
    #         "down": "JOY1_DOWN",
    #     }]
    #
    #     f.write("\n[input]\n")
    #
    #     for i, port in enumerate(self.ports):
    #         if not port.device:
    #             continue
    #         if i == 0:
    #             key = "joystick_port_1"
    #         else:
    #             key = "joystick_port_0"
    #         if port.type == "amiga_mouse":
    #             value = "mouse"
    #         else:
    #             value = port.device.id
    #         print("[INPUT] Configure device:", key, value)
    #         f.write("{0} = {1}\n".format(key, value))

    # def find_kickstart(self, bios_patterns):
    #     print("find_kickstart", bios_patterns)
    #     for name in os.listdir(self.firmware_dir):
    #         for bios in bios_patterns:
    #             if bios in name:
    #                 return os.path.join(self.firmware_dir, name)
    #     raise Exception("Could not find Amiga bios/kickstart " +
    #                     repr(bios_patterns))



import fsgs
from fsgs import openretro
from launcher.settings.maintenance_settings_page import DefragmentDatabasesTask

import fsboot
import fsui
from fsui.extra.taskdialog import TaskDialog
from launcher.option import Option
from launcher.res import gettext
from launcher.settings.settings_page import SettingsPage


class LauncherSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("fs-uae-launcher", "pkg:launcher")
        self.add_header(icon, "{} Launcher".format(fsgs.product))

        if fsboot.get("fws") == "1":
            # We omit the appearance settings, since they have no effect
            # when running under the workspace environment.
            pass
        else:
            self.add_option(Option.LAUNCHER_THEME)
            self.add_option(Option.LAUNCHER_FONT_SIZE)

        self.add_option(Option.LAUNCHER_CLOSE_BUTTONS)

        self.add_section(gettext("Experimental Features"))
        # Netplay feature is now enabled by default
        # self.add_option(Option.NETPLAY_FEATURE)
        if not openretro:
            self.add_option(Option.PLATFORMS_FEATURE)
        # self.add_option(Option.LAUNCHER_CONFIG_FEATURE)
        # self.add_option(Option.LAUNCHER_SETUP_WIZARD_FEATURE)

        self.add_section(gettext("Maintenance"))
        label = fsui.MultiLineLabel(
            self,
            gettext(
                "Defragmenting the databases will improve performance "
                "by ensuring that tables and indices are stored contiguously "
                "on disk. It will also reclaim some storage space."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_top=20)
        button = fsui.Button(self, gettext("Defragment Databases"))
        button.activated.connect(self.on_defragment_button)
        self.layout.add(button, margin_top=20)

    def on_defragment_button(self):
        TaskDialog(self.get_window(), DefragmentDatabasesTask()).show()



import os
import re
import unicodedata
import unittest
import itertools


QUOTED_TERMS_RE = re.compile('["].*?["]')


class GameNameUtil(object):

    TOSEC = "TOSEC"
    NOINTRO = "NOINTRO"

    @classmethod
    def get_letter(cls, name):
        letter_name = name.upper()
        if letter_name.startswith("THE "):
            letter_name = letter_name[4:]
        if letter_name.startswith("A "):
            letter_name = letter_name[2:]
        for i in range(len(letter_name)):
            letter = letter_name[i]
            if letter in "01234567890":
                letter = "0"
                break
            if letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                break
        else:
            raise Exception("could not find letter for game")
        return letter

    @classmethod
    def create_fs_name(cls, name):
        name = name.replace(":", " - ")
        name = name.replace("*", "-")
        name = name.replace("?", "")
        name = name.replace("/", "-")
        name = name.replace('"', "'")
        name = name.replace(">>>", " ")
        name = name.replace("<<<", " ")
        # FIXME: regular expression instead
        for i in range(3):
            name = name.replace("  ", " ")
        while name.endswith("."):
            name = name[:-1]
        name = name.strip()
        return name

    @classmethod
    def create_cmp_name(cls, name):
        name = str(name)
        decomposed = unicodedata.normalize("NFD", name)
        result = []
        for c in decomposed.lower():
            if c in "abcdefghijklmnopqrstuvwxyz0123456789+":
                result.append(c)
        return "".join(result)

    # noinspection SpellCheckingInspection
    @classmethod
    def create_cmpname(cls, name):
        return cls.create_cmp_name(name)

    @classmethod
    def create_link_name(cls, name):
        name = str(name)
        decomposed = unicodedata.normalize("NFD", name)
        result = []
        for c in decomposed.lower():
            if c in "abcdefghijklmnopqrstuvwxyz0123456789+-":
                pass
            else:
                c = "-"
            if c != "-" or not result or result[-1] != "-":
                result.append(c)
        if result and result[-1] == "-":
            result = result[:-1]
        return "".join(result)

    @classmethod
    def is_bad_dump(cls, path):
        if "[b]" in path:
            return True
        m = re.search("\\[b[0-9]? [^\\]]*\\]", path, re.IGNORECASE)
        if m is not None:
            return True
        return False

    @classmethod
    def extract_platform(cls, path):
        p = path.upper()
        n = os.path.basename(p)
        for alias, platform in GamePlatform.aliases.items():
            if "(" + alias + "," in n:
                return platform
            if "(" + alias + ")" in n:
                return platform
        while os.path.dirname(p) != p:
            p = os.path.dirname(p)
            n = os.path.basename(p)
            for alias, platform in GamePlatform.aliases.items():
                if "(" + alias + "," in n:
                    return platform
                if "(" + alias + ")" in n:
                    return platform
            if n in GamePlatform.aliases:
                return GamePlatform.aliases[n]
        print("extract_platform returning None for", path)
        return None

    @classmethod
    def find_base_name(cls, path):
        name = os.path.basename(path)
        # name = os.path.normcase(name)
        pos_list = [name.find("("), name.find("["), name.rfind("."), len(name)]
        for pos in sorted(pos_list):
            if pos > -1:
                return name[:pos].strip().lower()
        return name.strip().lower()

    @classmethod
    def find_alt(cls, path):
        m = re.search("\\[a([0-9]*)\\]", path, re.IGNORECASE)
        if m is not None:
            if not m.group(1):
                return 1
            return int(m.group(1))
        return None

    @classmethod
    def strip_alt(cls, path):
        path = re.sub("\\[a([0-9]*)\\]", "", path)
        return path

    @classmethod
    def find_disk_count(cls, name):
        for i in range(20):
            if "of {0})".format(i) in name:
                return i
        return 0

    @classmethod
    def find_number(cls, name):
        # find medium number, if any

        # number = None
        lname = name.lower()
        for prefix in "reel", "disk", "cd":
            # print(prefix)
            pprefix = "(" + prefix + " "
            index = lname.find(pprefix)
            # print(index)
            if index >= 0:
                rname = lname[index + len(pprefix) :].strip()
                # print(rname)
                num = ""
                onlydigits = True
                for c in rname:
                    if c in "0123456789":
                        num += c
                    elif c in "abcdefghijklmnopqrstuvwxyz":
                        onlydigits = False
                        num += c
                    else:
                        break
                if not num:
                    pass
                elif onlydigits:
                    # number = int(num)
                    return int(num)
                else:
                    num = num.lower()
                    if num == "i":
                        return 1
                    elif num == "ii":
                        return 2
                    elif num == "iii":
                        return 3
                    elif num == "iv":
                        return 4
                    elif num == "v":
                        return 5
                    elif num == "vi":
                        return 6
                    elif num == "vii":
                        return 7
                    elif num == "viii":
                        return 8
                    elif num == "ix":
                        return 9
                    elif num == "x":
                        return 10
                    elif len(num) == 1:
                        return 1 + ord(num) - ord("a")
                    else:
                        return num
                # print(' ----------- ', rname, '---', num, number)
        # return number

    @classmethod
    def strip_number(cls, name):
        # remove '(disk xxx)' from name
        # and also (disk xxx)(yyy)
        name = re.sub("\\([Dd][Ii][Ss][Kk][^)]*\\)(\\([^\\)]*\\))?", "", name)
        # remove 'disc x' from name
        name = re.sub("\\([Dd][Ii][Ss][Cc][^)]*\\)", "", name)
        # 'reel x' is used by Cinemaware games
        name = re.sub("\\([Rr][Ee][Ee][Ll][^)]*\\)", "", name)
        # remove 'cd x'
        name = re.sub("\\([Cc][Dd][^)]*\\)", "", name)
        # remove 'track x' from multi-part bin files
        name = re.sub("\\([Tt][Tr][Aa][Cc][Kk][^)]*\\)", "", name)
        # remove 'xxx disk' from name
        # name = re.sub('\\([^\\(]* [Dd][Ii][Ss][Kk]\\)', '', name)
        return name

    @classmethod
    def extract_disk_label(cls, name):
        # match '(disk xxx)' from name and also (disk xxx)(yyy)
        print(name)
        base_name = ""
        primary = ""

        m = re.search("\\([Dd][Ii][Ss][Kk]([^)]*)\\)(\\([^\\)]*\\))?", name)
        if m:
            base_name = "Disk "
        else:
            for i in range(30, -1, -1):
                n = "disk" + str(i)
                if n in name.lower().replace(" ", ""):
                    primary = "Disk " + str(i)
                    break

        if not primary and name.endswith(".ipf"):
            use_name = name.replace("Disc", "Disk")

            # if "ProgramDisc" in name:
            #     primary = "Program Disk"
            # if "ProgramDisk" in name:
            #     primary = "Program Disk"
            # elif "DataDisc" in name:
            #     primary = "Data Disk"
            # elif "DataDisk" in name:
            #     primary = "Data Disk"
            if not primary:
                for i in range(30, -1, -1):
                    n = "disk" + str(i)
                    if n in use_name.lower().replace(" ", ""):
                        primary = "Disk " + str(i)
                        break
            if not primary:
                fname, ext = os.path.splitext(use_name)
                label = fname.split("_")[-1]
                label = label.replace("Disk", " Disk ")
                label = label.replace("  ", " ").strip()
                primary = label

            # if not primary:
            #     raise Exception("unknown IPF name")

        # remove 'disc x' from name
        # name = re.sub('\\([Dd][Ii][Ss][Cc][^)]*\\)', '', name)
        #  'reel x' is used by Cinemaware games
        # name = re.sub('\\([Rr][Ee][Ee][Ll][^)]*\\)', '', name)
        #  remove 'cd x'
        # name = re.sub('\\([Cc][Dd][^)]*\\)', '', name)
        #  remove 'track x' from multi-part bin files
        # name = re.sub('\\([Tt][Tr][Aa][Cc][Kk][^)]*\\)', '', name)
        #  remove 'xxx disk' from name
        # name = re.sub('\\([^\\(]* [Dd][Ii][Ss][Kk]\\)', '', name)
        # return ""name""

        if not primary:
            primary = "Game Disk"
        secondary = ""
        if m:
            primary = m.group(1).strip()
            primary = primary.split("of")[0].strip()
            if m.group(2):
                # secondary = " (" + m.group(2).strip() + ")"
                secondary = " " + m.group(2).strip()
        else:
            secondary = ""
        # if secondary:
        return "{0}{1}{2}".format(base_name, primary, secondary)
        # return "{0}{1}".format(base_name, primary)

    @classmethod
    def strip_flags(cls, name):
        name = re.sub("\\[[^\\]]*\\]", "", name)
        # name = name.replace('(Boot)', '')
        return name

    # FIXME: Mark as deprecated
    @classmethod
    def extract_names(cls, path, info=None, style=TOSEC):
        return cls.extract_names_from_file_name(path, info=info, style=style)

    @classmethod
    def extract_names_from_file_name(cls, path, info=None, style=TOSEC):
        if info is None:
            info = {}
        # game_name = ""
        config_name = ""

        name = os.path.basename(path)
        # FIXME: no longer allowed to send in paths!
        if "/" in path or "\\" in path:
            print(
                "WARNING: slash (or backslash in extract_names arg:",
                repr(path),
            )

        # name = path
        # print(name)

        # find real case of file name

        # if os.path.dirname(path):
        #     for item in os.listdir(os.path.dirname(path)):
        #         if os.path.normcase(item) == os.path.normcase(name):
        #             name = item
        #             break
        name, ext = os.path.splitext(name)

        # def ireplace(s, search, replace):
        #     index = s.lower().find(search.lower())
        #     if index >= 0:
        #         return s[:index] + replace + s[index+len(search):]
        #     return s

        name = cls.strip_number(name)
        # name = ireplace(name, '(save disk)', '')
        # name = ireplace(name, '(intro disk)', '')

        # TOSEC naming convention places vx.yy right after name,
        # without paranthesis - put paranthesis around version

        # print(name)
        name = re.sub("( v[0-9A-Z][^ ]*) \\(", "(\\1) (", name)
        # also sometimes rXX where XX is the revision
        name = re.sub("( r[0-9A-Z][^ ]*) \\(", "(\\1) (", name)
        # print("...", name)

        # remove adjacent spaces
        name = re.sub("[ ]+", " ", name)

        name = name.replace("[", "(")
        name = name.replace("]", ")")

        parts = name.split("(")
        for i, part in enumerate(parts):
            if i == 0:
                continue
            if part[-1] != ")":
                continue
            if style == cls.TOSEC:
                comma_parts = part.split(", ")
                if len(comma_parts) == 2:
                    parts[i] = comma_parts[1][:-1] + " " + comma_parts[0] + ")"
            elif style == cls.NOINTRO:
                part = part.replace(",", ") (")
                part = re.sub("[ ]+", " ", part)
                parts[i] = part
            else:
                raise Exception("unknown style")

        name = "(".join(parts)

        # find end of name (and start of config name)
        pos = name.find("(")
        pos2 = name.find("[")
        if pos > 0:
            if 0 < pos2 < pos:
                pos = pos2
        if pos > 0:
            game_name = name[:pos].strip()
            config_name = name[pos:]
            config_name = config_name.replace("[!]", " ")
            config_name = config_name.replace(",", ", ")
            config_name = config_name.replace(")", ", ")
            # config_name = config_name.replace(']', ', ')
            config_name = config_name.replace("(", "")
            # config_name = config_name.replace('[', '')
            config_name = config_name.replace(" ,", ",")
            config_name = config_name.strip(" ,")
            config_name = re.sub("[ ]+", " ", config_name)
        else:
            game_name = name.strip()
        # if game_name.lower().endswith(', the'):
        #     game_name = 'The ' + game_name[:-5]
        #     game_name = game_name.strip()
        # elif game_name.lower().endswith(', a'):
        #     game_name = 'A ' + game_name[:-3]
        #     game_name = game_name.strip()
        return game_name, config_name

    @classmethod
    def extract_index_terms(cls, name, include_combinations=True, expand=True):
        print("extract search terms for:", name)

        name = name.replace("\u00df", "ss")  # convert eszett

        words = name.lower().split(" ")

        new_words = []
        for word in words:
            # if word.endswith("n'"):
            #     print("- replacing n' with ng")
            #     word = word[:-2] + "ng"
            print("word:", word)
            letters = [" "]
            for letter in word:
                decomposed = unicodedata.normalize("NFD", letter)
                if len(decomposed) > 1:
                    for c in decomposed:
                        if c in "abcdefghijklmnopqrstuvwxyz0123456789+":
                            letters.append(c)
                else:
                    c = decomposed
                    # we want to include original punctuation, for now

                    # letters.append(c)
                    if c in "abcdefghijklmnopqrstuvwxyz0123456789+":
                        letters.append(c)
                    else:
                        # punctuation is replaced with blanks
                        if not letters[-1] == " ":
                            letters.append(" ")
            new_word = "".join(letters).strip()
            print("- new_word:", new_word)
            sub_words = new_word.split(" ")
            print("- sub_words:", sub_words)
            if include_combinations:
                combinations = [sub_words]
                if len(sub_words) > 0:
                    for i in range(1, len(sub_words)):
                        combinations.extend(
                            itertools.combinations(sub_words, i)
                        )
                print("- combinations:")
                for combination in combinations:
                    print(" ", combination)
                    new_words.append("".join(combination))
            else:
                new_words.extend(sub_words)
        print("new words:", new_words)

        terms = set()
        for word in new_words:
            try:
                add_words = normalize_words[word]
                print(" - replacement", word, "=>", add_words)
            except:
                add_words = [word]

            for add_word in add_words:
                if add_word.endswith("n") and expand:
                    # leavin(') ->
                    terms.add(add_word + "g")
                else:
                    terms.add(add_word)

        terms.difference_update(["a", "the", "and", "of", "for", "in"])
        print("terms", terms)
        return terms

    @classmethod
    def extract_search_terms(cls, name):
        quoted_terms = []

        def callback(m):
            quoted_terms.append(m.group(0))
            return ""

        stripped_name = QUOTED_TERMS_RE.sub(callback, name)
        result = cls.extract_index_terms(
            stripped_name, include_combinations=False, expand=False
        )
        for term in quoted_terms:
            result.add(term)
        return result

    # @classmethod
    # def extract_search_terms(cls, name):
    #     words = name.lower().split(" ")
    #
    #     new_words = []
    #     for word in words:
    #         #if word.endswith("n'"):
    #         #    print("- replacing n' with ng")
    #         #    word = word[:-2] + "ng"
    #         print("word:", word)
    #         letters = [" "]
    #         for letter in word:
    #             decomposed = unicodedata.normalize('NFD', letter)
    #             if len(decomposed) > 1:
    #                 for c in decomposed:
    #                     if c in "abcdefghijklmnopqrstuvwxyz0123456789+":
    #                         letters.append(c)
    #             else:
    #                 c = decomposed
    #                 # we want to include original punctuation, for now
    #                 #letters.append(c)
    #                 if c in "abcdefghijklmnopqrstuvwxyz0123456789+":
    #                     letters.append(c)
    #                 else:
    #                     # punctuation is replaced with blanks
    #                     if not letters[-1] == " ":
    #                         letters.append(" ")
    #         new_word = "".join(letters).strip()
    #         print("- new_word:", new_word)
    #         sub_words = new_word.split(" ")
    #         #print("- sub_words:", sub_words)
    #         #combinations = [sub_words]
    #         #if len(sub_words) > 0:
    #         #    for i in range(1, len(sub_words)):
    #         #        combinations.extend(itertools.combinations(sub_words, i))
    #         #print("- combinations:")
    #         #for combination in combinations:
    #         #    print(" ", combination)
    #         #    new_words.append("".join(combination))
    #     print("new words:", new_words)
    #
    #     terms = set()
    #     for word in new_words:
    #         try:
    #             add_words = normalize_words[word]
    #             print(" - replacement", word, "=>", add_words)
    #         except:
    #             add_words = [word]
    #
    #         for add_word in add_words:
    #             #terms.add(add_word)
    #             if add_word.endswith("n"):
    #                 # leavin(') -> leaving
    #                 terms.add(add_word + "g")
    #     print("terms", terms)
    #     return terms


normalize_words = {
    "se": ["special", "edition"],
    "one": ["1"],
    "two": ["2"],
    "three": ["3"],
    "four": ["4"],
    "five": ["5"],
    "six": ["6"],
    "seven": ["7"],
    "eight": ["8"],
    "nine": ["9"],
    "ten": ["10"],
    "eleven": ["11"],
    "twelve": ["12"],
    "thirteen": ["13"],
    "fourteen": ["14"],
    "fifteen": ["15"],
    "sixteen": ["16"],
    "seventeen": ["17"],
    "eighteen": ["18"],
    "nineteen": ["19"],
    "twenty": ["20"],
    "i": ["1"],
    "ii": ["2"],
    "iii": ["3"],
    "iv": ["4"],
    "v": ["5"],
    "vi": ["6"],
    "vii": ["7"],
    "viii": ["8"],
    "ix": ["9"],
    "x": ["10"],
    "xi": ["11"],
    "xii": ["12"],
    "n": ["and"],
    "versus": ["vs"],
    "brothers": ["bros"],
    "doctor": ["dr"],
    "them": ["em"],
    "mister": ["mr"],
    "miss": ["ms"],
    "television": ["tv"],
}
for _year_i in range(1980, 2000):
    normalize_words[str(_year_i - 1900)] = [str(_year_i)]


class TestGameNameUtil(unittest.TestCase):
    def test_already_commaed_name(self):
        full_name = "Name (A, B, C).adf"
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Name")
        self.assertEquals(variant, "A, B, C")

    def test_variant_with_slash(self):
        full_name = "Tron (8/9).dummy"
        name, variant = GameNameUtil.extract_names(full_name)
        print("name:", name)
        print("variant:", variant)
        self.assertEquals(name, "Tron")
        self.assertEquals(variant, "8/9")

    def test_name_without_variant(self):
        full_name = "Name.adf"
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Name")
        self.assertEquals(variant, "")

    def test_tosec_name(self):
        full_name = (
            "Lotus Turbo Challenge 2 (1991)(Gremlin)[cr CPY]"
            "[t +4 Goonies].adf"
        )
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Lotus Turbo Challenge 2")
        self.assertEquals(variant, "1991, Gremlin, cr CPY, t +4 Goonies")

    def test_tosec_name_with_comma_in_publisher(self):
        full_name = "Zero Gravity (1997)(Hollemans, M.)(AGA).adf"
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Zero Gravity")
        self.assertEquals(variant, "1997, M. Hollemans, AGA")

    def test_tosec_name_with_comma_in_publisher_2(self):
        full_name = "Ziriax (1990)(Software Business, The)[3001].dummy"
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Ziriax")
        self.assertEquals(variant, "1990, The Software Business, 3001")

    def test_hitchhikers_guide_to_the_galaxy(self):
        full_name = "Leather Goddesses of Phobos r59 (1986)(Infocom).adf"
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Leather Goddesses of Phobos")
        self.assertEquals(variant, "r59, 1986, Infocom")

    def test_leather_goddesses_of_phobos(self):
        full_name = (
            "Hitchhiker's Guide to the Galaxy, The r58 " "(1986)(Infocom).adf"
        )
        name, variant = GameNameUtil.extract_names(full_name)
        self.assertEquals(name, "Hitchhiker's Guide to the Galaxy, The")
        self.assertEquals(variant, "r58, 1986, Infocom")

    def test_deep_core(self):
        full_name = "Deep Core (Europe)(v1.00).dummy"
        name, variant = GameNameUtil.extract_names_from_file_name(
            full_name, style=GameNameUtil.NOINTRO
        )
        self.assertEquals(name, "Deep Core")
        self.assertEquals(variant, "Europe, v1.00")

    def test_extract_index_terms_kings_quest(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("King's Quest IV"),
            set(["king", "kings", "s", "quest", "4"]),
        )

    def test_extract_index_terms_alien_breed_2(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Alien Breed II"),
            set(["alieng", "breed", "2"]),
        )

        self.assertEqual(
            GameNameUtil.extract_index_terms("Alien Breed 2"),
            set(["alieng", "breed", "2"]),
        )

    def test_extract_index_terms_project_x_se_92(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Project-X SE '92"),
            set(["projectx", "project", "10", "special", "editiong", "1992"]),
        )

    def test_extract_index_terms_leaving_teramis(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Leavin' Teramis"),
            set(["leaving", "teramis"]),
        )

    def test_extract_index_terms_fa_18_interceptor(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("F/A-18 Interceptor"),
            set(["f", "fa", "fa18", "f18", "a18", "f", "18", "interceptor"]),
        )

    def test_extract_index_sensible_soccer_95_96(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Sensible Soccer '95/'96"),
            set(["sensible", "soccer", "1995", "1996", "9596"]),
        )

        self.assertEqual(
            GameNameUtil.extract_index_terms("Sensible Soccer 1995-96"),
            set(["sensible", "soccer", "1995", "1996", "199596"]),
        )

    def test_extract_index_bump_n_burn(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Bump 'n' Burn"),
            set(["bump", "burng"]),
        )

    def test_extract_index_bump_n_burn_2(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Bump'n' Burn"),
            set(["bump", "burng", "bumpng"]),
        )

    def test_extract_index_bump_n_burn_3(self):
        self.assertEqual(
            GameNameUtil.extract_index_terms("Bump'n Burn"),
            set(["bump", "burng", "bumpng"]),
        )

    def test_search_bump_n_burn(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Bump 'n' Burn").issuperset(
                GameNameUtil.extract_search_terms("Bump'n Burn")
            )
        )

    def test_search_bump_n_burn_2(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Bump 'n' Burn").issuperset(
                GameNameUtil.extract_search_terms("Bump and Burn")
            )
        )

    def test_search_bump_n_burn_3(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms(
                "Lotus Turbo Challenge 2"
            ).issuperset(GameNameUtil.extract_search_terms("Lotus II"))
        )

    def test_search_sensible_1(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms(
                "Sensible Soccer '95/'96"
            ).issuperset(GameNameUtil.extract_search_terms("soccer 95/96"))
        )

    def test_search_fa_18(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("F/A-18 Interceptor").issuperset(
                GameNameUtil.extract_search_terms("fa 18")
            )
        )

    def test_search_formula_1(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Formula 1").issuperset(
                GameNameUtil.extract_search_terms("formula one")
            )
        )

    def test_search_formula_1_2(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Formula one").issuperset(
                GameNameUtil.extract_search_terms("formula 1")
            )
        )

    def test_search_projectx_1(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Project-X SE").issuperset(
                GameNameUtil.extract_search_terms("project x special")
            )
        )

    def test_search_projectx_2(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms(
                "Project-X Special Edition"
            ).issuperset(GameNameUtil.extract_search_terms("projectx se"))
        )

    def test_search_yo_joe(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Yo! Joe!").issuperset(
                GameNameUtil.extract_search_terms("yo joe")
            )
        )

    def test_search_alien3(self):
        self.assertTrue(
            GameNameUtil.extract_index_terms("Alien 3").issuperset(
                GameNameUtil.extract_search_terms("Alien\u00b3")
            )
        )


if __name__ == "__main__":
    unittest.main()



from fsbc.desktop import open_url_in_browser
from launcher.launcher_config import LauncherConfig
from fsui import Image, HorizontalLayout, Menu
from ...i18n import gettext
from .StatusElement import StatusElement


url_descriptions = [
    ("database_url", "Game Database Entry"),
    ("homepage_url", "Game Home Page"),
    ("longplay_url", "Longplay Video"),
    ("hol_url", "Hall of Light Entry"),
    ("lemon_url", "LemonAmiga Entry"),
    ("wikipedia_url", "Wikipedia Entry"),
    ("mobygames_url", "MobyGames Entry"),
    ("amigamemo_url", "AmigaMemo Entry"),
    ("thelegacy_url", "TheLegacy Entry"),
    ("whdload_url", "WHDLoad Slave Info"),
]

url_keys = set([x[0] for x in url_descriptions])


class WebLinkElement(StatusElement):
    def __init__(self, parent):
        StatusElement.__init__(self, parent)
        # self.set_min_width(140)
        self.layout = HorizontalLayout()
        self.icon = Image("launcher:res/16x16/world_link.png")
        self.right_icon = Image("launcher:res/16x16/drop_down_arrow.png")
        self.right_icon_disabled = Image(
            "launcher:res/16x16/drop_down_arrow_disabled.png"
        )
        # self.inactive_icon = self.active_icon.grey_scale()

        self.text = gettext("Web Links")

        LauncherConfig.add_listener(self)
        self.on_config("protection", LauncherConfig.get("protection"))

        self.have = set()
        for key in url_keys:
            self.on_config(key, LauncherConfig.get(key))

    def on_left_down(self):
        if len(self.have) == 0:
            return
        menu = Menu()
        for key, description in url_descriptions:
            value = LauncherConfig.get(key, "")
            if value:
                menu.add_item(description, create_open_url_function(value))
        self.popup_menu(menu)

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key in url_keys:
            if value:
                self.have.add(key)
            elif key in self.have:
                self.have.remove(key)
            old_active = self.active
            self.active = len(self.have) > 0
            if old_active != self.active:
                self.refresh()


def create_open_url_function(value):
    def open_url():
        open_url_in_browser(value)

    return open_url



#!/usr/bin/env python
# xdfscan
# quickly scan large sets of Amiga disk image files

from __future__ import absolute_import
from __future__ import print_function

import sys
import argparse
import os.path
import time

from amitools.fs.blkdev.BlkDevFactory import BlkDevFactory
from amitools.fs.validate.Validator import Validator
from amitools.fs.validate.Progress import Progress

# ----- logging -----

class MyProgress(Progress):
  def __init__(self):
    Progress.__init__(self)
    self.clk = int(time.clock() * 1000)
  def begin(self, msg):
    Progress.begin(self, msg)
  def add(self):
    Progress.add(self)
    clk = int(time.clock() * 1000)
    delta = clk - self.clk
    # update display every 250ms
    if delta > 250:
      self.clk = clk
      print("%s: %d\r" % (self.msg, self.num)),
      sys.stdout.flush()

def pre_log_path(path, msg):
  print("%20s  %s  \r" % (msg, path)),
  sys.stdout.flush()

def log_path(path, msg):
  print("%20s  %s  " % (msg, path))

def print_block(percent):
  print("%3.1f%%\r" % (percent / 10.0)),
  sys.stdout.flush()

# ----- scanner -----

factory = BlkDevFactory()

def scan(path, args):
  if not os.path.exists(path):
    log_path(path, "DOES NOT EXIST")
    return 1
  if os.path.isdir(path):
    return scan_dir(path, args)
  elif os.path.isfile(path):
    return scan_file(path, args)

def scan_dir(path, args):
  for name in sorted(os.listdir(path)):
    epath = os.path.join(path, name)
    result = scan(epath, args)
    if result != 0:
      return result
  return 0

def check_extension(path, args):
  ext = []
  if not args.skip_disks:
    ext += ['.adf','.adz','.adf.gz']
  if not args.skip_hds:
    ext += ['.hdf']
  for a in ext:
    if path.endswith(a):
      return True
  return False

def scan_file(path, args):
  if not check_extension(path, args):
    return 0
  try:
    pre_log_path(path,"scan")
    ret_code = 0
    ret_str = ""
    stay = True

    # create a block device for image file
    blkdev = factory.open(path, read_only=True)

    # create validator
    progress = MyProgress()
    v = Validator(blkdev, min_level=args.level, debug=args.debug, progress=progress)

    # 1. check boot block
    res = []
    boot_dos, bootable = v.scan_boot()
    if boot_dos:
      # 2. check root block
      root = v.scan_root()
      if not root:
        # disk is bootable
        if bootable:
          res.append("boot")
        else:
          res.append("    ")
        # invalid root
        res.append("nofs")
      else:
        # 3. scan tree
        v.scan_dir_tree()
        # 4. scan files
        v.scan_files()
        # 5. scan_bitmap
        v.scan_bitmap()

        # summary
        e, w = v.get_summary()
        if w > 0:
          res.append("w%03d" % w)
        if e > 0:
          res.append("E%03d" % e)
        else:
          res.append("    ")
        # disk is bootable
        if bootable:
          res.append("boot")
        else:
          res.append("    ")
        if e == 0 and w == 0:
          res.append(" ok ")
        else:
          res.append("NOK ")
    else:
      # boot block is not dos
      res.append("NDOS")

    # report result
    if len(res) == 0:
      res.append("done")
    log_path(path," ".join(res))
    # summary
    if args.verbose:
      v.log.dump()
    return ret_code
  except IOError as e:
    log_path(path,"BLKDEV?")
    if args.verbose:
      print(e)
    return 0

# ----- main -----
def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('input', nargs='+', help="input image file or directory (to scan tree)")
  parser.add_argument('-v', '--verbose', action='store_true', default=False, help="be more verbos")
  parser.add_argument('-d', '--debug', action='store_true', default=False, help="show debug info")
  parser.add_argument('-q', '--quick', action='store_true', default=False, help="quick mode. faster: skip image if root is invalid")
  parser.add_argument('-l', '--level', default=2, help="show only level or above (0=debug, 1=info, 2=warn, 3=error)", type=int)
  parser.add_argument('-D', '--skip-disks', action='store_true', default=False, help="do not scan disk images")
  parser.add_argument('-H', '--skip-hds', action='store_true', default=False, help="do not scan hard disk images")
  args = parser.parse_args()

  # main scan loop
  ret = 0
  for i in args.input:
    ret = scan(i, args)
    if ret != 0:
      break
  return ret


if __name__ == '__main__':
  try:
    sys.exit(main())
  except KeyboardInterrupt as e:
    print("aborting...")



from weakref import ref
from .qt import Qt, QDesktopWidget, QSignal

# from .. import default_window_center, default_window_parent


_windows = set()


# noinspection PyPep8Naming
def WindowBase(BaseClass):

    # noinspection PyPep8Naming
    class WindowBaseImplementation(BaseClass):

        closed = QSignal()

        def __init__(self, parent, *args, title="", border=True, **kwargs):
            # if parent is None and len(default_window_parent) > 0:
            #     parent = default_window_parent[-1]
            #     print("using default parent", parent)
            #     parent = parent.real_window()

            super().__init__(parent, *args, **kwargs)
            # MixinBase.__init__(self)

            # noinspection PyUnusedLocal
            # def init_window(self, parent, title):
            # self.init_mixin_base()
            self.setWindowTitle(title)

            self.layout = None
            self._size_specified = False
            self.already_closed = False
            self.close_listeners = []
            _windows.add(self)
            self.destroyed.connect(self.__destroyed)
            self.closed.connect(self.__closed)
            self._window = ref(self)
            self.setAttribute(Qt.WA_DeleteOnClose, True)
            if not border:
                self.setWindowFlags(
                    Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint
                )
                # self.setWindowFlags(Qt.FramelessWindowHint)
            self._centered_on_initial_show = False
            if hasattr(self, "accepted"):
                self.accepted.connect(self.__accepted)
            if hasattr(self, "rejected"):
                self.rejected.connect(self.__rejected)

        def center_on_initial_show(self):
            if self._centered_on_initial_show:
                return
            if self.layout and not self._size_specified:
                self.set_size(self.layout.get_min_size())
            self.on_resize()
            self.center_on_parent()
            self._centered_on_initial_show = True

        def __closed(self):
            print(str(self) + ".__closed")
            _windows.remove(self)

        def __destroyed(self):
            print(str(self) + ".__destroyed")

        def get_parent(self):
            return None

        def add_close_listener(self, function):
            # self.close_listeners.append(function)
            self.closed.connect(function)

        def get_window(self):
            return self

        def get_container(self):
            return self

        def set_icon(self, icon):
            self.setWindowIcon(icon.qicon())

        def set_icon_from_path(self, _):
            print("FIXME: Window.set_icon_from_path")

        def get_position(self):
            return self.pos().x(), self.pos().y()

        def set_position(self, position):
            self.move(position[0], position[1])

        def get_size(self):
            return self.width(), self.height()

        def set_size(self, size):
            self._size_specified = True
            # self.SetClientSize(size)
            # print("FIXME:\n\nDialog.set_size")
            self.resize(size[0], size[1])

        def get_title(self):
            return self.windowTitle()

        def set_title(self, title):
            self.setWindowTitle(title)

        def is_maximized(self):
            # return self.isMaximized()
            return self.windowState() == Qt.WindowMaximized
            # print("FIXME: always returning False")
            # return False

        def maximize(self, maximize=True):
            if maximize:
                self.showMaximized()
            else:
                self.setWindowState(Qt.WindowNoState)

        def minimize(self):
            self.setWindowState(Qt.WindowMinimized)

        def get_window_center(self):
            return self.x() + self.width() // 2, self.y() + self.height() // 2

        def center_on_parent(self):
            real_parent = self.parent()
            if real_parent:
                pp = real_parent.x(), real_parent.y()
                ps = real_parent.width(), real_parent.height()
                ss = self.get_size()
                self.move(
                    pp[0] + (ps[0] - ss[0]) // 2, pp[1] + (ps[1] - ss[1]) // 2
                )
            # elif len(default_window_center) > 0:
            #     x, y = default_window_center[-1]
            #     ss = self.get_size()
            #     self.move(x - ss[0] // 2, y - ss[1] // 2,)

        def center_on_screen(self):
            frect = self.frameGeometry()
            frect.moveCenter(QDesktopWidget().availableGeometry().center())
            self.move(frect.topLeft())

        def set_background_color(self, color):
            p = self.palette()
            p.setColor(self.backgroundRole(), color)
            self.setPalette(p)

        # def show(self):
        #     if hasattr(self, "layout") and not self._size_specified:
        #         self.set_size(self.layout.get_min_size())
        #     #QMainWindow.show(self)
        #     print("")
        #     print("")
        #     print(" -- show --")
        #     print("")
        #     print("")
        #     # noinspection PyUnresolvedReferences
        #     super().show()

        def is_shown(self):
            return self.isVisible()

        def showEvent(self, _):
            if self.layout and not self._size_specified:
                self.set_size(self.layout.get_min_size())
            self.on_resize()

        def raise_and_activate(self):
            self.raise_()
            self.activateWindow()

        def on_close(self):
            pass

        def closeEvent(self, event):
            print(str(self) + ".closeEvent")
            event.accept()
            self.__on_close()

        def __rejected(self):
            print(str(self) + ".__rejected")
            self.__on_close()

        def __accepted(self):
            print(str(self) + ".__accepted")
            self.__on_close()

        def __on_close(self):
            print("WindowMixin.cleanup_on_close_event")
            if self.already_closed:
                print("Looks like a duplicate event, ignoring this one")
                return

            self.closed.emit()
            self.on_close()

            self.already_closed = True

        def resizeEvent(self, _):
            self.on_resize()

        def on_resize(self):
            if self.layout:
                self.layout.set_size(self.get_size())
                self.layout.update()

        def close(self):
            # keep this method override, it is here so references to
            # object.close can be connected to signals
            super().close()

    return WindowBaseImplementation



"""
The goals of this class are:
1. To be able to use ISO-8859-1 encoding instead of CP43, and
2. Also support zip files with UTF-8 encoding, and

Python 3 assumes strings must be converted to CP437 for non-ASCII characters.
This is why there'll be decoding/encoding from/to CP437 even for ISO-8859-1
and UTF-8 name lookup (with Python 3). Complications arise because Python
seems to detect UTF-8 in some zip files (based on a flag) and then
automatically convert to/from unicode.
"""
import os
import sys
import zipfile
import zlib

try:
    from typing import List
except ImportError:
    from fstd.typing import List


class ZipFile(zipfile.ZipFile):
    def __init__(self, path: str, mode: str = "r") -> None:
        zipfile.ZipFile.__init__(self, path, mode)

    def getinfo(self, name: str) -> zipfile.ZipInfo:
        try:
            n = name.encode("ISO-8859-1").decode("CP437")
            return zipfile.ZipFile.getinfo(self, n)
        except Exception:
            pass
        try:
            n = name.encode("UTF-8").decode("CP437")
            return zipfile.ZipFile.getinfo(self, n)
        except Exception:
            pass
        return zipfile.ZipFile.getinfo(self, name)

    def namelist(self) -> List[str]:
        names = zipfile.ZipFile.namelist(self)
        for i, name in enumerate(names):
            print("...", repr(name))
            try:
                name = name.encode("CP437").decode("UTF-8")
            except Exception:
                try:
                    name = name.encode("CP437").decode("ISO-8859-1")
                except Exception:
                    pass
            names[i] = name
        return names


def _get_compressor(compress_type):
    if compress_type == zipfile.ZIP_DEFLATED:
        return zlib.compressobj(9, zlib.DEFLATED, -15)
    # noinspection PyProtectedMember,PyUnresolvedReferences
    return zipfile._get_compressor_fspy(compress_type)


def create_deterministic_archive(
    src, dst, fix_pyc_timestamps=False, torrentzip=False
):
    sz = ZipFile(src, "r")
    # Sort and remove duplicates
    names = sorted(set(sz.namelist()), key=str.lower)
    # FIXME: Remove non-needed empty directories
    dz = ZipFile(dst, "w")
    for name in names:
        data = sz.read(name)
        if name.endswith(".pyc") and fix_pyc_timestamps:
            data = data[0:4] + b"\x00\x00\x00\x00" + data[8:]
        zinfo = zipfile.ZipInfo()
        zinfo.filename = name
        zinfo.date_time = (1996, 12, 24, 23, 32, 00)
        zinfo.compress_type = zipfile.ZIP_DEFLATED
        zinfo.create_system = 0
        zinfo.create_version = 0
        zinfo.extract_version = 20
        zinfo.flag_bits = 2
        if torrentzip:
            zinfo.external_attr = 0
        else:
            zinfo.external_attr = (0o644 & 0xFFFF) << 16
            if name.endswith("/"):
                # FIXME: Check if this is correct?
                zinfo.external_attr |= 0x10
        try:
            if torrentzip:
                # noinspection PyProtectedMember
                zipfile._get_compressor_fspy = zipfile._get_compressor
                zipfile._get_compressor = _get_compressor
            dz.writestr(zinfo, data)
        finally:
            if torrentzip:
                # noinspection PyProtectedMember,PyUnresolvedReferences
                zipfile._get_compressor = zipfile._get_compressor_fspy
                # noinspection PyProtectedMember,PyUnresolvedReferences
                del zipfile._get_compressor_fspy
    sz.close()
    if torrentzip:
        # Placeholder for the CRC-32 of the central directory records.
        dz.comment = b"TORRENTZIPPED-XXXXXXXX"
    dz.close()
    if torrentzip:
        with open(dst, "r+b") as f:
            f.seek(-22 - 2 - 4 - 4, 2)
            size = (
                f.read(1)[0]
                + f.read(1)[0] * 256
                + f.read(1)[0] * 256 ** 2
                + f.read(1)[0] * 256 ** 2
            )
            socd = (
                f.read(1)[0]
                + f.read(1)[0] * 256
                + f.read(1)[0] * 256 ** 2
                + f.read(1)[0] * 256 ** 2
            )
            f.seek(socd)
            data = f.read(size)
            checksum = zlib.crc32(data)
            f.seek(-8, 2)
            f.write("{:X}".format(checksum).encode("ASCII"))


def convert_deterministic_archive(
    src, fix_pyc_timestamps=False, torrentzip=False
):
    tmp = os.path.join(
        os.path.dirname(src), ".~" + os.path.basename(src) + ".tmp"
    )
    try:
        create_deterministic_archive(
            src,
            tmp,
            fix_pyc_timestamps=fix_pyc_timestamps,
            torrentzip=torrentzip,
        )
        os.remove(src)
        os.rename(tmp, src)
    finally:
        if os.path.exists(tmp):
            os.remove(tmp)


def main():
    if len(sys.argv) > 1:
        if sys.argv[1] == "deterministic":
            fix_pyc_timestamps = False
            if "--fix-pyc-timestamps" in sys.argv:
                sys.argv.remove("--fix-pyc-timestamps")
                fix_pyc_timestamps = True
            torrentzip = False
            if "--torrentzip" in sys.argv:
                sys.argv.remove("--torrentzip")
                torrentzip = True
            if len(sys.argv) == 3:
                convert_deterministic_archive(
                    sys.argv[2],
                    torrentzip=torrentzip,
                    fix_pyc_timestamps=fix_pyc_timestamps,
                )
            else:
                create_deterministic_archive(
                    sys.argv[2],
                    sys.argv[3],
                    torrentzip=torrentzip,
                    fix_pyc_timestamps=fix_pyc_timestamps,
                )


if __name__ == "__main__":
    main()



from fsui.qt import QLineEdit
from fsui.qt.signal import Signal, SignalWrapper
from fsui.qt.widget_mixin import WidgetMixin


class TextField(QLineEdit, WidgetMixin):

    changed_signal = Signal()
    activated_signal = Signal()

    def __init__(self, parent, text="", read_only=False):
        QLineEdit.__init__(self, text, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        self.setReadOnly(read_only)
        # noinspection PyUnresolvedReferences
        self.textChanged.connect(self.__text_changed)
        # noinspection PyUnresolvedReferences
        self.returnPressed.connect(self.__return_pressed)

        self.changed = SignalWrapper(self, "changed")
        self.activated = SignalWrapper(self, "activated")

    def get_text(self):
        return self.text()

    def set_text(self, text):
        self.setText(text)

    def set_cursor_position(self, position):
        self.setCursorPosition(position)

    def on_changed(self):
        pass

    def __text_changed(self, _):
        self.changed.emit()
        self.on_changed()

    def select_all(self):
        self.selectAll()

    def __return_pressed(self):
        self.activated.emit()


class PasswordField(TextField):
    def __init__(self, parent, text=""):
        TextField.__init__(self, parent, text)
        self.setEchoMode(QLineEdit.Password)



#!/usr/bin/env python
# xdftool
# swiss army knife for adf and hdf amiga disk images

from __future__ import absolute_import
from __future__ import print_function

import sys
import argparse
import os.path

from amitools.fs.ADFSVolume import ADFSVolume
from amitools.fs.blkdev.BlkDevFactory import BlkDevFactory
from amitools.fs.FSError import *
from amitools.fs.Imager import Imager
from amitools.fs.Repacker import Repacker
from amitools.fs.block.BootBlock import BootBlock
from amitools.fs.block.RootBlock import RootBlock
from amitools.util.CommandQueue import CommandQueue
from amitools.util.HexDump import *
from amitools.util.DisAsm import DisAsm
import amitools.util.KeyValue as KeyValue
from amitools.fs.rdb.RDisk import RDisk
from amitools.fs.FSString import FSString
import amitools.fs.DosType as DosType

# system encoding
def make_fsstr(s):
  if sys.version_info[0] == 3:
    if isinstance(s, str):
      return FSString(s)
  # fetch default encoding (if available)
  encoding = sys.stdin.encoding
  if encoding is None:
    # assume win default encoding
    if os.platform == "win32":
      encoding = "cp1252"
    else:
      encoding = "utf-8"
  u = s.decode(encoding)
  return FSString(u)

# ----- commands -----
class Command:
  def __init__(self, args, opts, edit=False):
    self.args = args
    self.opts = opts
    self.edit = edit
    self.exit_code = 0

    self.volume = None
    self.blkdev = None

  def run(self, blkdev, vol):
    # optional init blkdev function
    if hasattr(self, "init_blkdev"):
      if blkdev == None:
        self.blkdev = self.init_blkdev(self.args.image_file)
        if self.blkdev == None:
          return 5
        blkdev = self.blkdev

    # optional init volume function
    if hasattr(self, "init_vol"):
      # close old
      if vol != None:
        vol.close()
      # create new volume
      self.volume = self.init_vol(blkdev)
      if self.volume == None:
        return 6
      vol = self.volume

    # common handler
    if hasattr(self, 'handle_blkdev'):
      return self.handle_blkdev(blkdev)
    elif hasattr(self, 'handle_vol'):
      return self.handle_vol(vol)
    else:
      return 0

  def has_init_blkdev(self):
    return hasattr(self, 'init_blkdev')

  def need_volume(self):
    return hasattr(self, 'handle_vol') and not hasattr(self, 'init_vol')

# ----- command handler -----
class FSCommandQueue(CommandQueue):
  def __init__(self, args, cmd_list, sep, cmd_map):
    CommandQueue.__init__(self, cmd_list, sep, cmd_map)
    self.args = args
    self.blkdev = None
    self.volume = None

  def run(self):
    self.img = self.args.image_file
    try:
      # main command loop
      exit_code = CommandQueue.run(self)
    except FSError as e:
      cmd = "'%s'" % " ".join(self.cmd_line)
      print(cmd, "FSError:", e)
      exit_code = 3
    except IOError as e:
      cmd = "'%s'" % " ".join(self.cmd_line)
      print(cmd, "IOError:", e)
      exit_code = 4
    finally:
      # close volume
      if self.volume != None:
        self.volume.close()
        if self.args.verbose:
          print("closing volume:", self.img)
      # close blkdev
      if self.blkdev != None:
        self.blkdev.close()
        if self.args.verbose:
          print("closing image:",self.img)
    return exit_code

  def create_cmd(self, cclass, name, opts):
    return cclass(self.args, opts)

  def _open_volume(self):
    # setup volume
    if self.volume == None:
      self.volume = ADFSVolume(self.blkdev)
      if self.args.verbose:
        print("opening volume:", self.img)
      self.volume.open()

  def run_first(self, cmd_line, cmd):
    self.cmd_line = cmd_line

    # check if first command is an init command
    if not cmd.has_init_blkdev():
      # auto add 'open' command
      pre_cmd = OpenCmd(self.args, [])
      if self.args.verbose:
        print("auto open command:", self.cmd_line)
      exit_code = pre_cmd.run(self.blkdev, self.volume)
      if self.args.verbose:
        print("auto open exit_code:", exit_code)
      if exit_code != 0:
        return exit_code
      self.blkdev = pre_cmd.blkdev
      # setup volume (if necessary)
      if cmd.need_volume():
        self._open_volume()

    # run first command
    if self.args.verbose:
      print("command:", self.cmd_line)
    if cmd.edit and self.args.read_only:
      raise IOError("Edit commands not allowed in read-only mode")

    # check code of command after __init__ parsing
    if cmd.exit_code != 0:
      return cmd.exit_code

    # perform command
    exit_code = cmd.run(self.blkdev, self.volume)
    if cmd.blkdev != None:
      self.blkdev = cmd.blkdev
    if cmd.volume != None:
      self.volume = cmd.volume

    # final exit code
    if self.args.verbose:
      print("exit_code:", exit_code)
    return exit_code

  def run_next(self, cmd_line, cmd):
    self.cmd_line = cmd_line
    if self.args.verbose:
      print("command:", self.cmd_line)
    # verify command
    if cmd.edit and self.args.read_only:
      raise IOError("Edit commands not allowed in read-only mode")
    # make sure volume is set up
    if self.volume == None and cmd.need_volume():
      self._open_volume()
    # run command
    exit_code = cmd.run(self.blkdev, self.volume)
    if cmd.blkdev != None:
      self.blkdev = cmd.blkdev
    if cmd.volume != None:
      self.volume = cmd.volume
    if self.args.verbose:
      print("exit_code:", exit_code)
    return exit_code

# ----- Init: Open/Create -----

class OpenCmd(Command):
  def init_blkdev(self, image_file):
    opts = KeyValue.parse_key_value_strings(self.opts)
    f = BlkDevFactory()
    return f.open(image_file, options=opts, read_only=self.args.read_only)

class CreateCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)

  def init_blkdev(self, image_file):
    opts = KeyValue.parse_key_value_strings(self.opts)
    f = BlkDevFactory()
    return f.create(image_file, options=opts, force=self.args.force)

class FormatCmd(Command):
  def init_blkdev(self, image_file):
    opts = KeyValue.parse_key_value_strings(self.opts[1:])
    f = BlkDevFactory()
    return f.create(image_file, options=opts, force=self.args.force)

  def init_vol(self, blkdev):
    vol = ADFSVolume(blkdev)
    n = len(self.opts)
    if n < 1 or n > 2:
      print("Usage: format <volume_name> [dos_type]")
      return None
    else:
      if n > 1:
        dos_str = self.opts[1]
        dos_type = DosType.parse_dos_type_str(dos_str)
        if dos_type is None:
          print("ERROR invalid dos_tpye:", dos_str)
          return None
      else:
        dos_type = None
      vol_name = make_fsstr(self.opts[0])
      vol.create(vol_name, dos_type=dos_type)
      return vol

# ----- Pack/Unpack -----

class PackCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
    self.imager = Imager()
    n = len(self.opts)
    if n == 0:
      print("Usage: pack <in_path> [dos_type] [out_size]")
      self.exit_code = 1
    else:
      self.in_path = self.opts[0]
      blkdev_opts = None
      dos_type = None
      if n > 1:
        # is a dostype given?
        dos_str = opts[1]
        dos_type = DosType.parse_dos_type_str(dos_str)
        if dos_type is not None:
          begin = 2
        else:
          begin = 1
        # take remainder as blkdev opts
        blkdev_opts = KeyValue.parse_key_value_strings(opts[begin:])
      self.blkdev_opts = blkdev_opts
      self.dos_type = dos_type
      self.imager.pack_begin(self.in_path)

  def init_blkdev(self, image_file):
    return self.imager.pack_create_blkdev(self.in_path, image_file, force=self.args.force, options=self.blkdev_opts)

  def init_vol(self, blkdev):
    return self.imager.pack_create_volume(self.in_path, blkdev, dos_type=self.dos_type)

  def handle_vol(self, volume):
    self.imager.pack_root(self.in_path, volume)
    self.imager.pack_end(self.in_path, volume)
    if self.args.verbose:
      print("Packed %d bytes" % (self.imager.get_total_bytes()))
    return 0

class UnpackCmd(Command):
  def handle_vol(self, vol):
    n = len(self.opts)
    if n == 0:
      print("Usage: unpack <out_path>")
      return 1
    else:
      out_path = self.opts[0]
      img = Imager()
      img.unpack(vol, out_path)
      if self.args.verbose:
        print("Unpacked %d bytes" % (img.get_total_bytes()))
      return 0

class RepackCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
    n = len(self.opts)
    if n == 0:
      print("Usage: repack <src_path> [in_size]")
      self.exit_code = 1
    in_img = self.opts[0]
    in_opts = KeyValue.parse_key_value_strings(self.opts[1:])
    self.repacker = Repacker(in_img, in_opts)
    if not self.repacker.create_in():
      self.exit_code = 2

  def init_blkdev(self, image_file):
    return self.repacker.create_out_blkdev(image_file)

  def init_vol(self, blkdev):
    return self.repacker.create_out_volume(blkdev)

  def handle_vol(self, vol):
    self.repacker.repack()
    return 0

# ----- Query Image -----

# list: list directory tree
class ListCmd(Command):
  def handle_vol(self, vol):
    n = len(self.opts)
    if n == 0:
      vol.root_dir.list(all=True)
      show_info = True
      show_all = True
      node = vol.get_root_dir()
    else:
      name = make_fsstr(self.opts[0])
      node = vol.get_path_name(name)
      if node != None:
        show_all = "all" in self.opts
        show_detail = "detail" in self.opts
        show_info = "info" in self.opts
        node.list(all=show_all, detail=show_detail)
      else:
        print("ERROR path not found:", node)
        return 2
    if show_info:
      info = node.get_info(show_all)
      for line in info:
        print(line)
    return 0

class TypeCmd(Command):
  def handle_vol(self, vol):
    p = self.opts
    if len(p) == 0:
      print("Usage: type <ami_file>")
      return 1
    else:
      name = make_fsstr(p[0])
      data = vol.read_file(name)
      print(data)
      return 0

class ReadCmd(Command):
  def handle_vol(self, vol):
    p = self.opts
    n = len(p)
    if n == 0 or n > 2:
      print("Usage: read <ami_file|dir> [sys_file]")
      return 1
    # determine output name
    out_name = os.path.basename(p[0])
    if n == 2:
      if os.path.isdir(p[1]):
        out_name = os.path.join(p[1],out_name)
      else:
        out_name = p[1]
    # single file operation
    name = make_fsstr(p[0])
    node = vol.get_path_name(name)
    if node == None:
      print("Node not found:", p[0])
      return 2
    # its a file
    if node.is_file():
      data = node.get_file_data()
      # write data to file
      fh = open(out_name,"wb")
      fh.write(data)
      fh.close()
    # its a dir
    elif node.is_dir():
      img = Imager()
      img.unpack_dir(node, out_name)
    node.flush()
    return 0

class InfoCmd(Command):
  def handle_vol(self, vol):
    info = vol.get_info()
    for line in info:
      print(line)
    return 0

# ----- Edit Image -----

class MakeDirCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    if len(self.opts) != 1:
      print("Usage: mkdir <dir_path>")
      return 1
    else:
      dir_path = make_fsstr(self.opts[0])
      vol.create_dir(dir_path)
      return 0

class WriteCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    n = len(self.opts)
    if n == 0 or n > 2:
      print("Usage: write <sys_file|dir> [ami_path]")
      return 1
    # get file_name and ami_path
    sys_file = self.opts[0]
    file_name = os.path.basename(sys_file)
    if n > 1:
        ami_path = self.opts[1]
    else:
        ami_path = os.path.basename(sys_file)
    # check sys path
    if not os.path.exists(sys_file):
      print("File not found:", sys_file)
      return 2

    ami_path = make_fsstr(ami_path)
    file_name = make_fsstr(file_name)
    # handle file
    if os.path.isfile(sys_file):
      fh = open(sys_file,"rb")
      data = fh.read()
      fh.close()
      vol.write_file(data, ami_path, file_name)
    # handle dir
    elif os.path.isdir(sys_file):
      parent_node, dir_name = vol.get_create_path_name(ami_path,file_name)
      if parent_node == None:
        print("Invalid path", ami_path)
        return 2
      node = parent_node.create_dir(dir_name)
      img = Imager()
      img.pack_dir(sys_file, node)

    return 0

class DeleteCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    n = len(self.opts)
    if n == 0:
      print("Usage: delete <ami_path> [wipe] [all]")
      return 1
    do_wipe = 'wipe' in self.opts
    do_all = 'all' in self.opts
    path = make_fsstr(self.opts[0])
    node = vol.delete(path, wipe=do_wipe, all=do_all)
    return 0

class ProtectCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    n = len(self.opts)
    if n != 2:
      print("Usage: protect <ami_file> <protect>")
      return 1
    name = make_fsstr(self.opts[0])
    pr_str = self.opts[1]
    node = vol.get_path_name(name)
    if node != None:
      node.change_protect_by_string(pr_str)
      return 0
    else:
      print("Can't find node:", name)
      return 2

class CommentCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    n = len(self.opts)
    if n != 2:
      print("Usage: comment <ami_file> <comment>")
      return 1
    name = make_fsstr(self.opts[0])
    comment = make_fsstr(self.opts[1])
    node = vol.get_path_name(name)
    if node != None:
      node.change_comment(comment)
      return 0
    else:
      print("Can't find node:", name)
      return 2

class TimeCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    n = len(self.opts)
    if n != 2:
      print("Usage: time <ami_file> <time>")
      return 1
    name = self.opts[0]
    tstr = self.opts[1]
    node = vol.get_path_name(name)
    if node != None:
      node.change_mod_ts_by_string(tstr)
      return 0
    else:
      print("Can't find node:", name)
      return 2

class RelabelCmd(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def handle_vol(self, vol):
    n = len(self.opts)
    if n != 1:
      print("Usage: relabel <new_name>")
      return 1
    name = self.opts[0]
    node = vol.relabel(name)
    return 0

# ----- Block Tools -----

class BlockCmd(Command):
  def handle_vol(self, vol):
    n = len(self.opts)
    if n == 0:
      print("Usage: block ( boot | root | node <ami_file> [data] | dump <block_no> )")
      return 1
    cmd = self.opts[0]
    if cmd == 'boot':
      vol.boot.dump()
      return 0
    elif cmd == 'root':
      vol.root.dump()
      return 0
    elif cmd == 'node':
      if n == 1:
        print("No node given!")
        return 1
      else:
        name = make_fsstr(self.opts[1])
        node = vol.get_path_name(name)
        if node != None:
          with_data = "data" in self.opts
          node.dump_blocks(with_data)
          return 0
        else:
          print("Can't find node:", name)
          return 2
    elif cmd == 'dump':
      if n == 1:
        print("No block number given!")
        return 1
      else:
        block_no = int(self.opts[1])
        data = vol.blkdev.read_block(block_no)
        print_hex(data)

# ----- Bitmap Tools -----

class BitmapCmd(Command):
  def handle_vol(self, vol):
    n = len(self.opts)
    if n == 0:
      print("Usage: bitmap ( free | used | find [n] | all | maps | root [all] | node <path> [all] [entries]) [brief]")
      return 1
    cmd = self.opts[0]

    # brief mode
    brief = False
    if self.opts[-1] == 'brief':
      brief = True
      self.opts = self.opts[:-1]

    if cmd == 'free':
      vol.bitmap.print_free(brief)
      return 0
    elif cmd == 'used':
      vol.bitmap.print_used(brief)
      return 0
    elif cmd == 'find':
      if n == 2:
        num = int(self.opts[1])
        blk_nums = vol.bitmap.find_n_free(num)
        if blk_nums == None:
          print("No %d free blocks found" % num)
          return 100
        else:
          print("Free %d blocks:" % num, blk_nums)
          return 0
      else:
        blk_num = vol.bitmap.find_free()
        if blk_num == None:
          print("No free block found")
          return 100
        else:
          print("Free block:", blk_num)
          return 0
    elif cmd == 'all':
      bm = vol.bitmap.create_draw_bitmap()
      vol.bitmap.draw_on_bitmap(bm)
      vol.root_dir.draw_on_bitmap(bm, True)
      vol.bitmap.print_draw_bitmap(bm, brief)
      return 0
    elif cmd == 'maps':
      bm = vol.bitmap.create_draw_bitmap()
      vol.bitmap.draw_on_bitmap(bm)
      vol.bitmap.print_draw_bitmap(bm, brief)
      return 0
    elif cmd == 'root':
      show_entries = 'entries' in self.opts
      bm = vol.bitmap.create_draw_bitmap()
      vol.root_dir.draw_on_bitmap(bm, False, show_entries)
      vol.bitmap.print_draw_bitmap(bm, brief)
      return 0
    elif cmd == 'node':
      if n > 1:
        name = make_fsstr(self.opts[1])
        node = vol.get_path_name(name)
        if node != None:
          show_all = 'all' in self.opts
          show_entries = 'entries' in self.opts
          bm = vol.bitmap.create_draw_bitmap()
          node.draw_on_bitmap(bm, show_all, show_entries)
          vol.bitmap.print_draw_bitmap(bm, brief)
          return 0
        else:
          print("Node '%s' not found!" % self.opts[1])
          return 2
      else:
        print("Need node path!")
        return 1
    else:
      print("Unknown bitmap command!")
      return 1

# ----- Root Tools -----

class RootCmd(Command):
  def handle_vol(self, volume):
    n = len(self.opts)
    if n == 0:
      print("Usage: root ( show | create_time <time> | disk_time <time> | time <time> )")
      return 1
    cmd = self.opts[0]
    # root show
    if cmd == 'show':
      volume.root.dump()
      return 0
    # create_time <time>
    elif cmd == 'create_time':
      if n != 2:
        print("create_time <time>")
        return 1
      else:
        volume.change_create_ts_by_string(self.opts[1])
        return 0
    # disk_time <time>
    elif cmd == 'disk_time':
      if n != 2:
        print("disk_time <time>")
        return 1
      else:
        volume.change_disk_ts_by_string(self.opts[1])
        return 0
    # time <time>
    elif cmd == 'time':
      if n != 2:
        print("time <time>")
        return 1
      else:
        volume.change_mod_ts_by_string(self.opts[1])
        return 0
    # boot ?
    else:
      print("Unknown root command!")
      return 1

# ----- Boot Tools -----

class BootCmd(Command):
  def handle_blkdev(self, blkdev):
    n = len(self.opts)
    if n == 0:
      print("Usage: boot ( show [hex] [asm] | read <file> | write <file> | install [boot1x] | clear )")
      return 1
    cmd = self.opts[0]
    # fetch boot blcok
    bb = BootBlock(blkdev, 0)
    bb.read()
    # boot show [hex] [asm]
    if cmd == 'show':
      bb.dump()
      if bb.valid and bb.boot_code != None:
        if 'hex' in self.opts:
          print_hex(bb.boot_code)
        if 'asm' in self.opts:
          dis = DisAsm()
          code = dis.disassemble(bb.boot_code)
          dis.dump(code)
      return 0
    # boot read <file>
    elif cmd == 'read':
      if n > 1:
        if bb.valid:
          if bb.boot_code != None:
            f = open(self.opts[1], "wb")
            f.write(bb.boot_code)
            f.close()
            return 0
          else:
            print("No Boot Code found!")
            return 2
        else:
          print("Invalid Boot Block!")
          return 1
      else:
        print("No file name!")
        return 1
    # boot write <file>
    elif cmd == 'write':
      if n > 1:
        f = open(self.opts[1], "rb")
        data = f.read()
        f.close()
        ok = bb.set_boot_code(data)
        if ok:
          bb.write()
          return 0
        else:
          print("Boot Code invalid!")
          return 2
      else:
        print("No file name!")
        return 1
    # boot install [<name>]
    elif cmd == 'install':
      if n == 1: # default boot code
        name = "boot2x3x"
      else:
        name = self.opts[1]
      # boot code directory
      bc_dir = bb.get_boot_code_dir()
      if bc_dir == None:
        print("No boot code directory found!")
        return 1
      path = os.path.join(bc_dir, name + ".bin")
      if not os.path.exists:
        print("Boot code '%s' not found!" % path)
        return 1
      # read data
      f = open(path,"rb")
      data = f.read()
      f.close()
      ok = bb.set_boot_code(data)
      if ok:
        bb.write()
        return 0
      else:
        print("Boot Code invalid!")
        return 2
    # boot clear
    elif cmd == 'clear':
      bb.set_boot_code(None)
      bb.write()
      return 0
    # boot ?
    else:
      print("Unknown boot command!")
      return 1

# ----- BlkDev Command -----

class BlkDevCmd(Command):
  def handle_blkdev(self, blkdev):
    blkdev.dump()
    return 0

# ----- main -----
def main():
  # call scanner and process all files with selected command
  cmd_map = {
  "open" : OpenCmd,
  "create" : CreateCmd,
  "list" : ListCmd,
  "type" : TypeCmd,
  "read" : ReadCmd,
  "makedir" : MakeDirCmd,
  "write" : WriteCmd,
  "delete" : DeleteCmd,
  "format" : FormatCmd,
  "bitmap" : BitmapCmd,
  "blkdev" : BlkDevCmd,
  "protect" : ProtectCmd,
  "comment" : CommentCmd,
  "time" : TimeCmd,
  "block" : BlockCmd,
  "pack" : PackCmd,
  "unpack" : UnpackCmd,
  "repack" : RepackCmd,
  "boot" : BootCmd,
  "root" : RootCmd,
  "info" : InfoCmd,
  "relabel" : RelabelCmd
  }

  parser = argparse.ArgumentParser()
  parser.add_argument('image_file')
  parser.add_argument('command_list', nargs='+', help="command: "+",".join(cmd_map.keys()))
  parser.add_argument('-v', '--verbose', action='store_true', default=False, help="be more verbos")
  parser.add_argument('-s', '--seperator', default='+', help="set the command separator char sequence")
  parser.add_argument('-r', '--read-only', action='store_true', default=False, help="read-only operation")
  parser.add_argument('-f', '--force', action='store_true', default=False, help="force overwrite existing image")
  args = parser.parse_args()

  cmd_list = args.command_list
  sep = args.seperator
  queue = FSCommandQueue(args, cmd_list, sep, cmd_map)
  code = queue.run()
  return code


if __name__ == '__main__':
  sys.exit(main())



import weakref


class Group(object):
    def __init__(self, parent):
        self._parent = weakref.ref(parent)
        if hasattr(parent, "_window"):
            # noinspection PyProtectedMember
            self._window = parent._window
        self.position = (0, 0)
        self.__visible = True
        # FIXME
        # parent.destroyed.connect(self.on_destroy)

    @property
    def parent(self):
        return self._parent()

    # def show_or_hide(self, show):
    #     self.__visible = show

    def is_visible(self):
        return self.__visible

    def on_destroy(self):
        pass

    def get_window(self):
        return self.parent.get_window()

    def get_container(self):
        return self.parent.get_container()

    def get_min_width(self):
        return self.layout.get_min_width()

    def get_min_height(self):
        return self.layout.get_min_height()

    def set_position(self, position):
        self.position = position
        if self.layout:
            self.layout.set_position(position)

    def set_size(self, size):
        if self.layout:
            self.layout.set_size(size)

    def set_position_and_size(self, position, size):
        self.position = position
        if self.layout:
            self.layout.set_position_and_size(position, size)



from functools import lru_cache

from fsgs.util import sdl2
from fsbc.system import macosx
from fsui.qt import Qt, QEvent


class KeySym:
    def __init__(self):
        self.sym = 0


class Key:
    def __init__(self):
        self.keysym = KeySym()


class Event:
    def __init__(self):
        self.type = 0
        self.key = None

    @classmethod
    def create_key_event(cls, ev):
        # event = Event()
        # event.key = Key()
        event = {}
        if ev.type() == QEvent.KeyPress:
            # event.type = sdl2.SDL_KEYDOWN
            event["type"] = "key-down"
        elif ev.type() == QEvent.KeyRelease:
            # event.type = sdl2.SDL_KEYUP
            event["type"] = "key-up"
        else:
            raise Exception("Unexpected event type in create_key_event")
        # key_map = get_key_map()
        # print(ev.key(), "vs", Qt.Key_Shift)
        # event.key.keysym.sym = key_map.get(ev.key(), 0)
        # event["key"] = key_map.get(ev.key(), 0)
        event["key"] = get_key(ev)
        return event

    @classmethod
    def create_mouse_event(cls, ev, window_size):
        event = {}
        if ev.type() == QEvent.MouseMove:
            event["type"] = "mouse-motion"
        elif ev.type() == QEvent.MouseButtonPress:
            event["type"] = "mouse-press"
        elif ev.type() == QEvent.MouseButtonRelease:
            event["type"] = "mouse-release"
        elif ev.type() == QEvent.MouseButtonDblClick:
            event["type"] = "mouse-press"
        else:
            raise Exception("Unexpected event type in create_mouse_event")
        # Normalize coordinates to a 1920x1080 OpenGL-like coordinate system
        x, y = ev.x(), ev.y()
        x = (x * 1920) // window_size[0]
        y = 1080 - (y * 1080) // window_size[1]
        event["pos"] = x, y
        return event

    @classmethod
    def create_fake_mouse_event(cls, type_, x, y, window_size):
        event = {"type": type_}
        x = (x * 1920) // window_size[0]
        y = 1080 - (y * 1080) // window_size[1]
        event["pos"] = x, y
        return event


def get_key(ev):
    print("Qt key:", ev.key(), int(ev.modifiers()), Qt.KeypadModifier)
    if macosx:
        # FIXME: TODO: CHECK
        # if ev.key() == Qt.Key_Meta:
        #     print("Control key, native virtual code:", ev.nativeVirtualCode())
        # elif ev.key() == Qt.Key_Shift:
        pass
    else:
        if ev.key() == Qt.Key_Control:
            print("Control key, native scan code:", ev.nativeScanCode())
            if ev.nativeScanCode() == 105:  # Linux
                return sdl2.SDLK_RCTRL
            # FIXME: TODO: WINDOWS
            return sdl2.SDLK_LCTRL
        elif ev.key() == Qt.Key_Shift:
            print("Shift key, native scan code:", ev.nativeScanCode())
            if ev.nativeScanCode() == 62:  # Linux
                return sdl2.SDLK_RSHIFT
            # FIXME: TODO: WINDOWS
            return sdl2.SDLK_LSHIFT
    # On OS X, the KeypadModifier value will also be set when an arrow
    # key is pressed as the arrow keys are considered part of the keypad.
    # http://doc.qt.io/qt-5/qt.html#KeyboardModifier-enum
    macos_arrow_key = macosx and ev.key() in [
        Qt.Key_Left,
        Qt.Key_Right,
        Qt.Key_Up,
        Qt.Key_Down,
    ]
    if int(ev.modifiers()) & Qt.KeypadModifier and not macos_arrow_key:
        print("keypad!")
        print(ev.key(), "vs", Qt.Key_4)
        return {
            Qt.Key_Insert: sdl2.SDLK_KP_0,
            Qt.Key_End: sdl2.SDLK_KP_1,
            Qt.Key_Down: sdl2.SDLK_KP_2,
            Qt.Key_PageDown: sdl2.SDLK_KP_3,
            Qt.Key_Left: sdl2.SDLK_KP_4,
            Qt.Key_Clear: sdl2.SDLK_KP_5,
            Qt.Key_Right: sdl2.SDLK_KP_6,
            Qt.Key_Home: sdl2.SDLK_KP_7,
            Qt.Key_Up: sdl2.SDLK_KP_8,
            Qt.Key_PageUp: sdl2.SDLK_KP_9,
            Qt.Key_Delete: sdl2.SDLK_KP_PERIOD,
        }.get(ev.key(), 0)
    key_map = get_key_map()
    return key_map.get(ev.key(), 0)


@lru_cache()
def get_key_map():
    result = {
        Qt.Key_Escape: sdl2.SDLK_ESCAPE,
        Qt.Key_Tab: sdl2.SDLK_TAB,
        # Qt.Key_Backtab	0x01000002
        Qt.Key_Backspace: sdl2.SDLK_BACKSPACE,
        Qt.Key_Return: sdl2.SDLK_RETURN,
        # Typically located on the keypad.
        Qt.Key_Enter: sdl2.SDLK_KP_ENTER,
        Qt.Key_Insert: sdl2.SDLK_INSERT,
        Qt.Key_Delete: sdl2.SDLK_DELETE,
        # The Pause/Break key (Note: Not anything to do with pausing media).
        Qt.Key_Pause: sdl2.SDLK_PAUSE,
        # Qt.Key_Print	0x01000009
        # Qt.Key_SysReq	0x0100000a
        # Qt.Key_Clear	0x0100000b
        Qt.Key_Home: sdl2.SDLK_HOME,
        Qt.Key_End: sdl2.SDLK_END,
        Qt.Key_Left: sdl2.SDLK_LEFT,
        Qt.Key_Up: sdl2.SDLK_UP,
        Qt.Key_Right: sdl2.SDLK_RIGHT,
        Qt.Key_Down: sdl2.SDLK_DOWN,
        Qt.Key_PageUp: sdl2.SDLK_PAGEUP,
        Qt.Key_PageDown: sdl2.SDLK_PAGEDOWN,
        # FIXME: LSHIFT vs RSHIFT
        Qt.Key_Shift: sdl2.SDLK_LSHIFT,
        # On Mac OS X, this corresponds to the Command keys.
        # Qt.Key_Control	0x01000021
        # FIXME: LCTRL vs RCTRL, OS X, etc
        Qt.Key_Control: sdl2.SDLK_LCTRL,
        # On Mac OS X, this corresponds to the Control keys. On Windows
        # keyboards, this key is mapped to the Windows key.
        # Qt.Key_Meta	0x01000022
        # Qt.Key_Alt	0x01000023
        # On Windows, when the KeyDown event for this key is sent, the
        # Ctrl+Alt modifiers are also set.
        # Qt.Key_AltGr	0x01001103
        # Qt.Key_CapsLock	0x01000024
        # Qt.Key_NumLock	0x01000025
        # Qt.Key_ScrollLock	0x01000026
        Qt.Key_F1: sdl2.SDLK_F1,
        Qt.Key_F2: sdl2.SDLK_F2,
        Qt.Key_F3: sdl2.SDLK_F3,
        Qt.Key_F4: sdl2.SDLK_F4,
        Qt.Key_F5: sdl2.SDLK_F5,
        Qt.Key_F6: sdl2.SDLK_F6,
        Qt.Key_F7: sdl2.SDLK_F7,
        Qt.Key_F8: sdl2.SDLK_F8,
        Qt.Key_F9: sdl2.SDLK_F9,
        Qt.Key_F10: sdl2.SDLK_F10,
        # Qt.Key_F11	0x0100003a
        # Qt.Key_F12	0x0100003b
        # Qt.Key_F13	0x0100003c
        # Qt.Key_F14	0x0100003d
        # Qt.Key_F15	0x0100003e
        # Qt.Key_F16	0x0100003f
        # Qt.Key_F17	0x01000040
        # Qt.Key_F18	0x01000041
        # Qt.Key_F19	0x01000042
        # Qt.Key_F20	0x01000043
        # Qt.Key_F21	0x01000044
        # Qt.Key_F22	0x01000045
        # Qt.Key_F23	0x01000046
        # Qt.Key_F24	0x01000047
        # Qt.Key_F25	0x01000048
        # Qt.Key_F26	0x01000049
        # Qt.Key_F27	0x0100004a
        # Qt.Key_F28	0x0100004b
        # Qt.Key_F29	0x0100004c
        # Qt.Key_F30	0x0100004d
        # Qt.Key_F31	0x0100004e
        # Qt.Key_F32	0x0100004f
        # Qt.Key_F33	0x01000050
        # Qt.Key_F34	0x01000051
        # Qt.Key_F35	0x01000052
        # Qt.Key_Super_L	0x01000053
        # Qt.Key_Super_R	0x01000054
        # Qt.Key_Menu	0x01000055
        # Qt.Key_Hyper_L	0x01000056
        # Qt.Key_Hyper_R	0x01000057
        # Qt.Key_Help	0x01000058
        # Qt.Key_Direction_L	0x01000059
        # Qt.Key_Direction_R	0x01000060
        Qt.Key_Space: sdl2.SDLK_SPACE,
        # Qt.Key_Any	Key_Space
        # Qt.Key_Exclam	0x21
        # Qt.Key_QuoteDbl	0x22
        # Qt.Key_NumberSign	0x23
        # Qt.Key_Dollar	0x24
        # Qt.Key_Percent	0x25
        # Qt.Key_Ampersand	0x26
        # Qt.Key_Apostrophe	0x27
        # Qt.Key_ParenLeft	0x28
        # Qt.Key_ParenRight	0x29
        # Qt.Key_Asterisk	0x2a
        # Qt.Key_Plus	0x2b
        # Qt.Key_Comma	0x2c
        # Qt.Key_Minus	0x2d
        # Qt.Key_Period	0x2e
        # Qt.Key_Slash	0x2f
        # Qt.Key_0	0x30
        # Qt.Key_1	0x31
        # Qt.Key_2	0x32
        # Qt.Key_3	0x33
        # Qt.Key_4	0x34
        # Qt.Key_5	0x35
        # Qt.Key_6	0x36
        # Qt.Key_7	0x37
        # Qt.Key_8	0x38
        # Qt.Key_9	0x39
        # Qt.Key_Colon	0x3a
        # Qt.Key_Semicolon	0x3b
        # Qt.Key_Less	0x3c
        # Qt.Key_Equal	0x3d
        # Qt.Key_Greater	0x3e
        # Qt.Key_Question	0x3f
        # Qt.Key_At	0x40
        Qt.Key_A: sdl2.SDLK_a,
        Qt.Key_B: sdl2.SDLK_b,
        Qt.Key_C: sdl2.SDLK_c,
        Qt.Key_D: sdl2.SDLK_d,
        Qt.Key_E: sdl2.SDLK_e,
        Qt.Key_F: sdl2.SDLK_f,
        Qt.Key_G: sdl2.SDLK_g,
        Qt.Key_H: sdl2.SDLK_h,
        Qt.Key_I: sdl2.SDLK_i,
        Qt.Key_J: sdl2.SDLK_j,
        Qt.Key_K: sdl2.SDLK_k,
        Qt.Key_L: sdl2.SDLK_l,
        Qt.Key_M: sdl2.SDLK_m,
        Qt.Key_N: sdl2.SDLK_n,
        Qt.Key_O: sdl2.SDLK_o,
        Qt.Key_P: sdl2.SDLK_p,
        Qt.Key_Q: sdl2.SDLK_q,
        Qt.Key_R: sdl2.SDLK_r,
        Qt.Key_S: sdl2.SDLK_s,
        Qt.Key_T: sdl2.SDLK_t,
        Qt.Key_U: sdl2.SDLK_u,
        Qt.Key_V: sdl2.SDLK_v,
        Qt.Key_W: sdl2.SDLK_w,
        Qt.Key_X: sdl2.SDLK_x,
        Qt.Key_Y: sdl2.SDLK_y,
        Qt.Key_Z: sdl2.SDLK_z,
        # FIXME: Depends on keymap?
        Qt.Key_BracketLeft: sdl2.SDLK_LEFTBRACKET,
        # Qt.Key_Backslash	0x5c
        # FIXME: Depends on keymap?
        Qt.Key_BracketRight: sdl2.SDLK_RIGHTBRACKET,
        # Qt.Key_BracketRight	0x5d
        # Qt.Key_AsciiCircum	0x5e
        # Qt.Key_Underscore	0x5f
        # Qt.Key_QuoteLeft	0x60
        # Qt.Key_BraceLeft	0x7b
        # Qt.Key_Bar	0x7c
        # Qt.Key_BraceRight	0x7d
        # Qt.Key_AsciiTilde	0x7e
        # Qt.Key_nobreakspace	0x0a0
        # Qt.Key_exclamdown	0x0a1
        # Qt.Key_cent	0x0a2
        # Qt.Key_sterling	0x0a3
        # Qt.Key_currency	0x0a4
        # Qt.Key_yen	0x0a5
        # Qt.Key_brokenbar	0x0a6
        # Qt.Key_section	0x0a7
        # Qt.Key_diaeresis	0x0a8
        # Qt.Key_copyright	0x0a9
        # Qt.Key_ordfeminine	0x0aa
        # Qt.Key_guillemotleft	0x0ab
        # Qt.Key_notsign	0x0ac
        # Qt.Key_hyphen	0x0ad
        # Qt.Key_registered	0x0ae
        # Qt.Key_macron	0x0af
        # Qt.Key_degree	0x0b0
        # Qt.Key_plusminus	0x0b1
        # Qt.Key_twosuperior	0x0b2
        # Qt.Key_threesuperior	0x0b3
        # Qt.Key_acute	0x0b4
        # Qt.Key_mu	0x0b5
        # Qt.Key_paragraph	0x0b6
        # Qt.Key_periodcentered	0x0b7
        # Qt.Key_cedilla	0x0b8
        # Qt.Key_onesuperior	0x0b9
        # Qt.Key_masculine	0x0ba
        # Qt.Key_guillemotright	0x0bb
        # Qt.Key_onequarter	0x0bc
        # Qt.Key_onehalf	0x0bd
        # Qt.Key_threequarters	0x0be
        # Qt.Key_questiondown	0x0bf
        # Qt.Key_Agrave	0x0c0
        # Qt.Key_Aacute	0x0c1
        # Qt.Key_Acircumflex	0x0c2
        # Qt.Key_Atilde	0x0c3
        # Qt.Key_Adiaeresis	0x0c4
        # Qt.Key_Aring	0x0c5
        # Qt.Key_AE	0x0c6
        # Qt.Key_Ccedilla	0x0c7
        # Qt.Key_Egrave	0x0c8
        # Qt.Key_Eacute	0x0c9
        # Qt.Key_Ecircumflex	0x0ca
        # Qt.Key_Ediaeresis	0x0cb
        # Qt.Key_Igrave	0x0cc
        # Qt.Key_Iacute	0x0cd
        # Qt.Key_Icircumflex	0x0ce
        # Qt.Key_Idiaeresis	0x0cf
        # Qt.Key_ETH	0x0d0
        # Qt.Key_Ntilde	0x0d1
        # Qt.Key_Ograve	0x0d2
        # Qt.Key_Oacute	0x0d3
        # Qt.Key_Ocircumflex	0x0d4
        # Qt.Key_Otilde	0x0d5
        # Qt.Key_Odiaeresis	0x0d6
        # Qt.Key_multiply	0x0d7
        # Qt.Key_Ooblique	0x0d8
        # Qt.Key_Ugrave	0x0d9
        # Qt.Key_Uacute	0x0da
        # Qt.Key_Ucircumflex	0x0db
        # Qt.Key_Udiaeresis	0x0dc
        # Qt.Key_Yacute	0x0dd
        # Qt.Key_THORN	0x0de
        # Qt.Key_ssharp	0x0df
        # Qt.Key_division	0x0f7
        # Qt.Key_ydiaeresis	0x0ff
        # Qt.Key_Multi_key	0x01001120
        # Qt.Key_Codeinput	0x01001137
        # Qt.Key_SingleCandidate	0x0100113c
        # Qt.Key_MultipleCandidate	0x0100113d
        # Qt.Key_PreviousCandidate	0x0100113e
        # Qt.Key_Mode_switch	0x0100117e
        # Qt.Key_Kanji	0x01001121
        # Qt.Key_Muhenkan	0x01001122
        # Qt.Key_Henkan	0x01001123
        # Qt.Key_Romaji	0x01001124
        # Qt.Key_Hiragana	0x01001125
        # Qt.Key_Katakana	0x01001126
        # Qt.Key_Hiragana_Katakana	0x01001127
        # Qt.Key_Zenkaku	0x01001128
        # Qt.Key_Hankaku	0x01001129
        # Qt.Key_Zenkaku_Hankaku	0x0100112a
        # Qt.Key_Touroku	0x0100112b
        # Qt.Key_Massyo	0x0100112c
        # Qt.Key_Kana_Lock	0x0100112d
        # Qt.Key_Kana_Shift	0x0100112e
        # Qt.Key_Eisu_Shift	0x0100112f
        # Qt.Key_Eisu_toggle	0x01001130
        # Qt.Key_Hangul	0x01001131
        # Qt.Key_Hangul_Start	0x01001132
        # Qt.Key_Hangul_End	0x01001133
        # Qt.Key_Hangul_Hanja	0x01001134
        # Qt.Key_Hangul_Jamo	0x01001135
        # Qt.Key_Hangul_Romaja	0x01001136
        # Qt.Key_Hangul_Jeonja	0x01001138
        # Qt.Key_Hangul_Banja	0x01001139
        # Qt.Key_Hangul_PreHanja	0x0100113a
        # Qt.Key_Hangul_PostHanja	0x0100113b
        # Qt.Key_Hangul_Special	0x0100113f
        # Qt.Key_Dead_Grave	0x01001250
        # Qt.Key_Dead_Acute	0x01001251
        # Qt.Key_Dead_Circumflex	0x01001252
        # Qt.Key_Dead_Tilde	0x01001253
        # Qt.Key_Dead_Macron	0x01001254
        # Qt.Key_Dead_Breve	0x01001255
        # Qt.Key_Dead_Abovedot	0x01001256
        # Qt.Key_Dead_Diaeresis	0x01001257
        # Qt.Key_Dead_Abovering	0x01001258
        # Qt.Key_Dead_Doubleacute	0x01001259
        # Qt.Key_Dead_Caron	0x0100125a
        # Qt.Key_Dead_Cedilla	0x0100125b
        # Qt.Key_Dead_Ogonek	0x0100125c
        # Qt.Key_Dead_Iota	0x0100125d
        # Qt.Key_Dead_Voiced_Sound	0x0100125e
        # Qt.Key_Dead_Semivoiced_Sound	0x0100125f
        # Qt.Key_Dead_Belowdot	0x01001260
        # Qt.Key_Dead_Hook	0x01001261
        # Qt.Key_Dead_Horn	0x01001262
        # Qt.Key_Back	0x01000061
        # Qt.Key_Forward	0x01000062
        # Qt.Key_Stop	0x01000063
        # Qt.Key_Refresh	0x01000064
        # Qt.Key_VolumeDown	0x01000070
        # Qt.Key_VolumeMute	0x01000071
        # Qt.Key_VolumeUp	0x01000072
        # Qt.Key_BassBoost	0x01000073
        # Qt.Key_BassUp	0x01000074
        # Qt.Key_BassDown	0x01000075
        # Qt.Key_TrebleUp	0x01000076
        # Qt.Key_TrebleDown	0x01000077
        # A key setting the state of the media player to play
        # Qt.Key_MediaPlay	0x01000080
        # A key setting the state of the media player to stop
        # Qt.Key_MediaStop	0x01000081
        # Qt.Key_MediaPrevious	0x01000082
        # Qt.Key_MediaNext	0x01000083
        # Qt.Key_MediaRecord	0x01000084
        # A key setting the state of the media player to pause
        # (Note: not the pause/break key)
        # Qt.Key_MediaPause	0x1000085
        # A key to toggle the play/pause state in the media player
        # (rather than setting an absolute state)
        # Qt.Key_MediaTogglePlayPause	0x1000086
        # Qt.Key_HomePage	0x01000090
        # Qt.Key_Favorites	0x01000091
        # Qt.Key_Search	0x01000092
        # Qt.Key_Standby	0x01000093
        # Qt.Key_OpenUrl	0x01000094
        # Qt.Key_LaunchMail	0x010000a0
        # Qt.Key_LaunchMedia	0x010000a1
        # On X11 this key is mapped to "My Computer" (XF86XK_MyComputer) key
        # for legacy reasons.
        # Qt.Key_Launch0	0x010000a2
        # On X11 this key is mapped to "Calculator" (XF86XK_Calculator) key
        # for legacy reasons.
        # Qt.Key_Launch1	0x010000a3
        # On X11 this key is mapped to XF86XK_Launch0 key for legacy reasons.
        # Qt.Key_Launch2	0x010000a4
        # Qt.Key_Launch3	0x010000a5	On X11 this key is mapped to
        # XF86XK_Launch1 key for legacy reasons.
        # Qt.Key_Launch4	0x010000a6	On X11 this key is mapped to
        # XF86XK_Launch2 key for legacy reasons.
        # Qt.Key_Launch5	0x010000a7	On X11 this key is mapped to
        # XF86XK_Launch3 key for legacy reasons.
        # Qt.Key_Launch6	0x010000a8	On X11 this key is mapped to
        # XF86XK_Launch4 key for legacy reasons.
        # Qt.Key_Launch7	0x010000a9	On X11 this key is mapped to
        # XF86XK_Launch5 key for legacy reasons.
        # Qt.Key_Launch8	0x010000aa	On X11 this key is mapped to
        # XF86XK_Launch6 key for legacy reasons.
        # Qt.Key_Launch9	0x010000ab	On X11 this key is mapped to
        # XF86XK_Launch7 key for legacy reasons.
        # Qt.Key_LaunchA	0x010000ac	On X11 this key is mapped to
        # XF86XK_Launch8 key for legacy reasons.
        # Qt.Key_LaunchB	0x010000ad	On X11 this key is mapped to
        # XF86XK_Launch9 key for legacy reasons.
        # Qt.Key_LaunchC	0x010000ae	On X11 this key is mapped to
        # XF86XK_LaunchA key for legacy reasons.
        # Qt.Key_LaunchD	0x010000af	On X11 this key is mapped to
        # XF86XK_LaunchB key for legacy reasons.
        # Qt.Key_LaunchE	0x010000b0	On X11 this key is mapped to
        # XF86XK_LaunchC key for legacy reasons.
        # Qt.Key_LaunchF	0x010000b1	On X11 this key is mapped to
        # XF86XK_LaunchD key for legacy reasons.
        # Qt.Key_LaunchG	0x0100010e	On X11 this key is mapped to
        # XF86XK_LaunchE key for legacy reasons.
        # Qt.Key_LaunchH	0x0100010f	On X11 this key is mapped to
        # XF86XK_LaunchF key for legacy reasons.
        # Qt.Key_MonBrightnessUp	0x010000b2
        # Qt.Key_MonBrightnessDown	0x010000b3
        # Qt.Key_KeyboardLightOnOff	0x010000b4
        # Qt.Key_KeyboardBrightnessUp	0x010000b5
        # Qt.Key_KeyboardBrightnessDown	0x010000b6
        # Qt.Key_PowerOff	0x010000b7
        # Qt.Key_WakeUp	0x010000b8
        # Qt.Key_Eject	0x010000b9
        # Qt.Key_ScreenSaver	0x010000ba
        # Qt.Key_WWW	0x010000bb
        # Qt.Key_Memo	0x010000bc
        # Qt.Key_LightBulb	0x010000bd
        # Qt.Key_Shop	0x010000be
        # Qt.Key_History	0x010000bf
        # Qt.Key_AddFavorite	0x010000c0
        # Qt.Key_HotLinks	0x010000c1
        # Qt.Key_BrightnessAdjust	0x010000c2
        # Qt.Key_Finance	0x010000c3
        # Qt.Key_Community	0x010000c4
        # Qt.Key_AudioRewind	0x010000c5
        # Qt.Key_BackForward	0x010000c6
        # Qt.Key_ApplicationLeft	0x010000c7
        # Qt.Key_ApplicationRight	0x010000c8
        # Qt.Key_Book	0x010000c9
        # Qt.Key_CD	0x010000ca
        # On X11 this key is not mapped for legacy reasons.
        # Use Qt.Key_Launch1 instead.
        # Qt.Key_Calculator	0x010000cb
        # Qt.Key_ToDoList	0x010000cc
        # Qt.Key_ClearGrab	0x010000cd
        # Qt.Key_Close	0x010000ce
        # Qt.Key_Copy	0x010000cf
        # Qt.Key_Cut	0x010000d0
        # Qt.Key_Display	0x010000d1
        # Qt.Key_DOS	0x010000d2
        # Qt.Key_Documents	0x010000d3
        # Qt.Key_Excel	0x010000d4
        # Qt.Key_Explorer	0x010000d5
        # Qt.Key_Game	0x010000d6
        # Qt.Key_Go	0x010000d7
        # Qt.Key_iTouch	0x010000d8
        # Qt.Key_LogOff	0x010000d9
        # Qt.Key_Market	0x010000da
        # Qt.Key_Meeting	0x010000db
        # Qt.Key_MenuKB	0x010000dc
        # Qt.Key_MenuPB	0x010000dd
        # Qt.Key_MySites	0x010000de
        # Qt.Key_News	0x010000df
        # Qt.Key_OfficeHome	0x010000e0
        # Qt.Key_Option	0x010000e1
        # Qt.Key_Paste	0x010000e2
        # Qt.Key_Phone	0x010000e3
        # Qt.Key_Calendar	0x010000e4
        # Qt.Key_Reply	0x010000e5
        # Qt.Key_Reload	0x010000e6
        # Qt.Key_RotateWindows	0x010000e7
        # Qt.Key_RotationPB	0x010000e8
        # Qt.Key_RotationKB	0x010000e9
        # Qt.Key_Save	0x010000ea
        # Qt.Key_Send	0x010000eb
        # Qt.Key_Spell	0x010000ec
        # Qt.Key_SplitScreen	0x010000ed
        # Qt.Key_Support	0x010000ee
        # Qt.Key_TaskPane	0x010000ef
        # Qt.Key_Terminal	0x010000f0
        # Qt.Key_Tools	0x010000f1
        # Qt.Key_Travel	0x010000f2
        # Qt.Key_Video	0x010000f3
        # Qt.Key_Word	0x010000f4
        # Qt.Key_Xfer	0x010000f5
        # Qt.Key_ZoomIn	0x010000f6
        # Qt.Key_ZoomOut	0x010000f7
        # Qt.Key_Away	0x010000f8
        # Qt.Key_Messenger	0x010000f9
        # Qt.Key_WebCam	0x010000fa
        # Qt.Key_MailForward	0x010000fb
        # Qt.Key_Pictures	0x010000fc
        # Qt.Key_Music	0x010000fd
        # Qt.Key_Battery	0x010000fe
        # Qt.Key_Bluetooth	0x010000ff
        # Qt.Key_WLAN	0x01000100
        # Qt.Key_UWB	0x01000101
        # Qt.Key_AudioForward	0x01000102
        # Qt.Key_AudioRepeat	0x01000103
        # Qt.Key_AudioRandomPlay	0x01000104
        # Qt.Key_Subtitle	0x01000105
        # Qt.Key_AudioCycleTrack	0x01000106
        # Qt.Key_Time	0x01000107
        # Qt.Key_Hibernate	0x01000108
        # Qt.Key_View	0x01000109
        # Qt.Key_TopMenu	0x0100010a
        # Qt.Key_PowerDown	0x0100010b
        # Qt.Key_Suspend	0x0100010c
        # Qt.Key_ContrastAdjust	0x0100010d
        # Qt.Key_MediaLast	0x0100ffff
        # Qt.Key_unknown	0x01ffffff
        # A key to answer or initiate a call (see Qt.Key_ToggleCallHangup
        # for a key to toggle current call state)
        # Qt.Key_Call	0x01100004
        # A key to activate the camera shutter
        # Qt.Key_Camera	0x01100020
        # Qt.Key_CameraFocus	0x01100021	A key to focus the camera
        # Qt.Key_Context1	0x01100000
        # Qt.Key_Context2	0x01100001
        # Qt.Key_Context3	0x01100002
        # Qt.Key_Context4	0x01100003
        # Qt.Key_Flip	0x01100006
        # A key to end an ongoing call (see Qt.Key_ToggleCallHangup for a
        # key to toggle current call state)
        # Qt.Key_Hangup	0x01100005
        # Qt.Key_No	0x01010002
        # Qt.Key_Select	0x01010000
        # Qt.Key_Yes	0x01010001
        # A key to toggle the current call state (ie. either answer, or
        # hangup) depending on current call state
        # Qt.Key_ToggleCallHangup	0x01100007
        # Qt.Key_VoiceDial	0x01100008
        # Qt.Key_LastNumberRedial	0x01100009
        # Qt.Key_Execute	0x01020003
        # Qt.Key_Printer	0x01020002
        # Qt.Key_Play	0x01020005
        # Qt.Key_Sleep	0x01020004
        # Qt.Key_Zoom	0x01020006
        # Qt.Key_Cancel	0x01020001
    }
    return result



import socket
import threading
import time
import traceback

from fsbc.application import call_after
from ..launcher_config import LauncherConfig
from ..launcher_signal import LauncherSignal


class ConnectionTester:
    def __init__(self):
        self.host_info = ("", 0)
        self.running = False
        self.stopping = False
        self.connection_attempt = 0

        LauncherConfig.add_listener(self)
        LauncherSignal.add_listener("quit", self)

    def on_quit_signal(self):
        print("on_quit_signal")
        self.stopping = True

    def on_config(self, key, _):
        if key in ["__netplay_addresses", "__netplay_port"]:
            addresses = LauncherConfig.get("__netplay_addresses")
            port = LauncherConfig.get("__netplay_port")
            host_info = (addresses, port)
            if host_info != self.host_info:
                self.host_info = host_info
            if not self.running:
                threading.Thread(
                    target=self.thread_main, name="ConnectionTesterThread"
                ).start()
                self.running = True

    def thread_main(self):
        try:
            self._thread_main()
        except Exception:
            traceback.print_exc()

    def set_host(self, host, last_error):
        if last_error:
            last_error = "attempt {1} {0}".format(
                last_error, self.connection_attempt
            )
        values = [
            ("__netplay_host", host),
            ("__netplay_host_last_error", last_error),
        ]
        if last_error or not host:
            # force not ready when we cannot connect to server
            values.append(("__netplay_ready", "0"))

        def function():
            LauncherConfig.set_multiple(values)

        call_after(function)

    def _thread_main(self):
        last_host_info = None
        last_check_time = 0
        while not self.stopping:
            t = time.time()
            host_info = self.host_info
            time_diff = t - last_check_time
            time_for_new_test = time_diff > 30.0 or host_info != last_host_info
            if host_info[0] and host_info[1] and time_for_new_test:
                last_check_time = t
                last_host_info = host_info
                try:
                    result, error = self.test_connection(host_info)
                    self.set_host(result, error)
                except Exception as e:
                    traceback.print_exc()
                    # FIXME: handle problem...
                    self.set_host("", repr(e))
            time.sleep(1.0)

    def test_connection(self, host_info):
        addresses, port = host_info
        port = int(port)
        addresses = addresses.split(",")
        verified_host = ""
        last_error = ""
        for host in addresses:
            self.connection_attempt += 1
            host = host.strip()
            s = None
            try:
                try:
                    print("trying", host, port)
                    s = socket.create_connection((host, port))
                    s.send(b"PING")
                    # FIXME: set timeout, don't want to wait forever here...
                    data = s.recv(4)
                    if data == b"PONG":
                        verified_host = host
                        last_error = ""
                        break
                    else:
                        last_error = "Wrong PING response"
                except Exception:
                    last_error = traceback.format_exc()
                    print(last_error)
            finally:
                if s is not None:
                    try:
                        s.close()
                    except Exception:
                        pass
        print("test connection - verified host:", verified_host)
        return verified_host, last_error



import warnings

import fsui.qt
from .widget_mixin import WidgetMixin
from fsui.qt.widget import Widget
from fsui.qt import QParent, QPushButton, QSignal


class ImageButton(Widget):

    activated = QSignal()

    def __init__(self, parent, image):
        super().__init__(parent)
        self.set_widget(QPushButton(QParent(parent)))
        fsui.qt.QPushButton.__init__(self, parent.get_container())
        icon = image.qicon
        self._widget.setIcon(icon)
        self._widget.setIconSize(fsui.qt.QSize(image.size[0], image.size[1]))
        self._widget.clicked.connect(self.__clicked)

    def set_image(self, image):
        self._widget.setIcon(image.qicon)

    def __clicked(self):
        self.on_activated()

    def on_activated(self):
        self.activated.emit()
        # FIXME: Remove
        self.on_activate()

    def on_activate(self):
        pass

    @property
    def clicked(self):
        warnings.warn("use activated instead", DeprecationWarning)
        return self.activated
